'''
Copyright (c) 2022 Masayuki TAKAHASHI

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), 
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

def AdditionalAnalysis(args):
    """
    Make fragment size matrix from new template sequence and the result containing primer sets that you have already generated.

    Parameters
    ----------
    args: ArgumentParser, class containing appropriate property as below
        input_file: Input file path (required)\n
        csv_file: File path of the CSV data obtained from other analysis (DesignIdentifyPrimer, DesignUniversalPrimer. required)\n
        output: Output directory path (Make a new directory if the directory does not exist)\n
        size_limit: The upper limit of amplicon size\n
        process: The number of processes (sometimes the number of CPU core) used for analysis\n
        forward: Forward primer sequence (required if you don't provide a CSV file with the '-f' option)\n
        reverse: Reverse primer sequence (required if you don't provide a CSV file with the '-f' option)\n
        circularDNA: If there are some circular DNAs in the input sequences, use this option (default: n/a. It means all input sequences are linear DNA. When there are some circular DNA input sequences, type 'all', 'individually', 'n/a', or the file path of the text that specify which sequence is circularDNA, after the '--circularDNA' option. See Readme for more detailed information.)\n
            Text file example:\n
                Sequence_name1 circularDNA\n
                Sequence_name2 linearDNA\n
                Sequence_name3 linearDNA\n
                    ...\n
                Sequence_nameN circularDNA\n
        warning: Shows all warnings when you use this option.

    Returns
    -------
    CSV_file: CSV file will be generated
    """

    import sys
    import os
    import re
    import warnings
    from functools import partial
    sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
    from tqdm import tqdm
    import math
    import csv
    import itertools as it
    import numpy as np
    import pandas as pd
    from multiprocessing import Pool, get_context, cpu_count
    from shrslib.basicfunc import init_worker, read_sequence_file, circularDNA, check_input_file, complementary_sequence
    from shrslib.scores import fragment_size_distance, calculate_score
    from shrslib.explore import PCR_amplicon
    from shrslib.multiprocessfunc import add_anal_worker_process, PCR_amplicon_with_progress_bar, cumulative_pairwise_identity
    try:
        input_file = args.input_file
    except:
        sys.exit("Need a FASTA file or folder containing FASTA files after '-i '. \nPlease use '-h' or '--help' option if you would like to confirm usage.")
    if os.path.isfile(input_file):
        file_paths = [input_file]
    elif os.path.isdir(input_file):
        file_names = os.listdir(input_file)
        if input_file.endswith("/"):
            file_paths = [input_file+file_name for file_name in file_names]
        else:
            file_paths = [input_file+"/"+file_name for file_name in file_names]
        file_paths = [file_path for file_path in file_paths if os.path.isfile(file_path)]
        for file_path in file_paths:
            with open(file_path, "rt", encoding = "utf-8") as fin:
                Firstline = fin.readline()
                if np.logical_not(Firstline.startswith(">") | Firstline.startswith("LOCUS")):
                    print("There are some file(s) of which are not FASTA or Genbank in folder inputed. Please remove the file(s).")
                    sys.exit()
    else:
        sys.exit("Error: The input file/folder does not exist.")
    if args.output is not None:
        output_folder = args.output
        if not output_folder.endswith("/"):
            output_folder = output_folder+"/"
    else:
        output_folder = ""
    isSummary_data = False
    Score_calculation = 'Fragment'
    if args.csv_file is not None:
        csv_file_path = args.csv_file
        csv_file_path = re.sub("\\\\", "/", os.path.abspath(csv_file_path))
        try:
            test_sequence = pd.read_csv(csv_file_path, header = 0)
            isSummary_data = True if ('Arguments' in test_sequence.columns) else isSummary_data
            isSummary_data = True if ('Arguments' in test_sequence.iloc[0].tolist()) else isSummary_data
            if 'Score_calculation_mode' in test_sequence.iloc[:, 0].tolist():
                Score_calculation = 'Sequence' if str(test_sequence.iloc[test_sequence.iloc[:, 0].tolist().index('Score_calculation_mode'), 1]) == 'Sequence' else Score_calculation
        except:
            with open(csv_file_path, "r", encoding = "utf-8") as fin:
                Read_Data = csv.reader(fin)
                max_col = np.max([len(line) for line in Read_Data])
            test_sequence = pd.read_csv(csv_file_path, names = range(max_col), dtype = object)
            isSummary_data = True if ('Arguments' in test_sequence.columns) else isSummary_data
            isSummary_data = True if ('Arguments' in test_sequence.iloc[0].tolist()) else isSummary_data
            if 'Score_calculation_mode' in test_sequence.iloc[:, 0].tolist():
                Score_calculation = 'Sequence' if str(test_sequence.iloc[test_sequence.iloc[:, 0].tolist().index('Score_calculation_mode'), 1]) == 'Sequence' else Score_calculation
            test_sequence = test_sequence.drop(0)
        if np.any(test_sequence.iloc[:, 2:].isnull().all(axis = 1)):
            test_sequence = test_sequence[~test_sequence.iloc[:, 2:].isnull().all(axis = 1)]
            test_sequence.columns = test_sequence.iloc[0]
            test_sequence.columns.name = None
            test_sequence = test_sequence.drop(test_sequence.index[0])
            test_sequence = test_sequence.drop(['No', 'Flexibility', 'Score', 'Fragment_number', 'Forward_Tm_value', 'Reverse_Tm_value', 'Tm_value_difference'], errors='ignore', axis = 1)
        fwd = list(test_sequence.iloc[:, 0])
        rev = list(test_sequence.iloc[:, 1])
        Check_File_Format = (np.sum(test_sequence.iloc[:, 0].apply(lambda x:len(re.sub("[ATGCBDHKMRSVWYNXatgcbdhkmrsvwynx]", "", str(x))))) == 0) & (np.sum(test_sequence.iloc[:, 1].apply(lambda x:len(re.sub("[ATGCBDHKMRSVWYNXatgcbdhkmrsvwynx]", "", str(x))))) == 0)
        Check_File_Format = Check_File_Format & test_sequence.iloc[:, 2:].applymap(lambda x:(str(x).replace(".", "").replace("-", "").isnumeric()) | (str(x).startswith("["))).all().all()
        Check_File_Format = Check_File_Format | test_sequence.iloc[:, 2:].applymap(lambda x:len(re.sub("[ATGCBDHKMRSVWYNXatgcbdhkmrsvwynx]", "", str(x))) == 0).all(axis = 1).any()
        if np.logical_not(Check_File_Format):
            print("Unexpected data format. Check CSV file.")
            sys.exit()
        if isSummary_data:
            isUniversal = ((np.sum(np.sum(test_sequence.iloc[:, 2:].applymap(lambda x:str(x).find("[") >= 0))) / test_sequence.iloc[:, 2:].size) < 1/6) & ((np.sum(np.sum(test_sequence.iloc[:, 2:].applymap(lambda x:len(re.sub("[ATGCBDHKMRSVWYNXatgcbdhkmrsvwynx]", "", str(x))) == 0))) / test_sequence.iloc[:, 2:].size) < 1/5)
            Dataset_type = 'Universal' if isUniversal else 'Identify'
        else:
            Dataset_type = 'Unknown'
    else:
        if args.forward is not None:
            fwd = [args.forward]
        else:
            sys.exit("Need forward primer or sequence sets csv file.")
        if args.reverse is not None:
            rev = [args.reverse]
        else:
            sys.exit("Need reverse primer or sequence sets csv file.")
        csv_file_path = ''
        Dataset_type = 'Unknown'
        isSummary_data = False
        test_sequence = pd.DataFrame([[fwd, rev]])
    amplicon_size_limit = args.size_limit
    CPU = args.process
    if CPU is not None:
        if CPU > cpu_count():
            CPU = int(math.ceil(cpu_count()))
    else:
        if cpu_count() <= 3:
            CPU = int(1)
        elif cpu_count() <= 8:
            CPU = int(math.floor(cpu_count() / 2))
        elif cpu_count() <= 12:
            CPU = int(cpu_count() - 4)
        elif cpu_count() <= 24:
            CPU = int(math.floor(cpu_count() / 2) + 2)
        else:
            CPU = int(16)
    Seq_no = []
    for file_path in file_paths:
        with open(file_path, "rt", encoding = "utf-8") as fin:
            read_FASTA = fin.read()
            Seq_no += [read_FASTA.count(">")]
    if args.circularDNA is not None:
        overlap_region = amplicon_size_limit
        if len(args.circularDNA) == 0:
            circular_index = check_input_file(file_paths, circular = "all")
        elif args.circularDNA[0].lower() == "all":
            circular_index = check_input_file(file_paths, circular = "all")
        elif args.circularDNA[0].lower() == "individually":
            circular_index = check_input_file(file_paths, circular = "individually")
        elif os.path.isfile(args.circularDNA[0]):
            circular_index = check_input_file(file_paths, circular = args.circularDNA[0])
        elif args.circularDNA[0].lower() == "n/a":
            overlap_region = 0
            circular_index = check_input_file(file_paths, circular = "n/a")
        else:
            print(" Specify the file path, 'all', 'individually' or 'n/a' after --circularDNA option.\n All input sequence(s) are analysed as circular DNA")
            circular_index = check_input_file(file_paths, circular = "all")
    else:
        overlap_region = 0
        circular_index = check_input_file(file_paths, circular = "n/a")
    Warning_ignore = args.warning
    if Warning_ignore:
        warnings.simplefilter('ignore')
    if ((np.all([type(f) is str for f in fwd])) & (np.all([type(r) is str for r in rev]))):
        Fragment_matrix = pd.DataFrame([])
        if len(fwd) > len(Seq_no):
            with tqdm(total = np.sum(Seq_no) * len(fwd), desc = "Now analyzing ...", leave = False, unit = "tasks", unit_scale = True) as pbar:
                def Progressbar_update(Result):
                    pbar.update(Result[1])
                for file_path in file_paths:
                    seq_dict = read_sequence_file(file_path)[0]
                    Sequence_length_list = [np.min([amplicon_size_limit, seq_dict[key].sequence_length]) for key in seq_dict]
                    seq_dict = {key:circularDNA(sequence = seq_dict[key], overlap = int(np.min([overlap_region, seq_dict[key].sequence_length]))) if circular_index[key] else circularDNA(sequence = seq_dict[key], overlap = int(0)) for key in seq_dict}
                    if (CPU > 1):
                        for limit, name in zip(Sequence_length_list, seq_dict):
                            with get_context("spawn").Pool(processes = min([CPU, len(fwd)]), initializer = init_worker) as pl:
                                try:
                                    Amplicons = [pl.apply_async(PCR_amplicon_with_progress_bar, kwds = {"forward": f, "reverse": r, "template": seq_dict[name], "Single_amplicon": False, "Sequence_Only": True, "amplicon_size_limit": limit, "allowance": 0, "Warning_ignore": Warning_ignore}, callback = Progressbar_update) for f, r in zip(fwd, rev)]
                                except KeyboardInterrupt:
                                    pl.terminate()
                                    pl.join()
                                    pl.close()
                                    print("\n\n --- Keyboard Interrupt ---")
                                    sys.exit("")
                                Amplicons = [amp.get()[0] for amp in Amplicons]
                            if Score_calculation == 'Sequence':
                                Amplicons = [amps if amps is not None else [] for amps in Amplicons]
                                Amplicons = [amps[0] if len(amps) == 1 else np.nan for amps in Amplicons]
                            else:
                                Amplicons = [[len(amp) for amp in amps] if amps is not None else np.nan for amps in Amplicons]
                            Amplicons = pd.DataFrame(pd.Series(Amplicons, index = pd.MultiIndex.from_tuples([mid for mid in zip(fwd, rev)]), name = name))
                            Fragment_matrix = pd.concat([Fragment_matrix, Amplicons], axis = 1, join = "outer")
                    else:
                        for limit, name in zip(Sequence_length_list, seq_dict):
                            Amplicons = [(PCR_amplicon(forward = f, reverse = r, template = seq_dict[name], Single_amplicon = False, Sequence_Only = True, amplicon_size_limit = limit, allowance = int(0), Warning_ignore = Warning_ignore), pbar.update(1), ) for f, r in zip(fwd, rev)]
                            Amplicons = [amp[0] for amp in Amplicons]
                            if Score_calculation == 'Sequence':
                                Amplicons = [amps if amps is not None else [] for amps in Amplicons]
                                Amplicons = [amps[0] if len(amps) == 1 else np.nan for amps in Amplicons]
                            else:
                                Amplicons = [[len(amp) for amp in amps] if amps is not None else np.nan for amps in Amplicons]
                            Amplicons = pd.DataFrame(pd.Series(Amplicons, index = pd.MultiIndex.from_tuples([mid for mid in zip(fwd, rev)]), name = name))
                            Fragment_matrix = pd.concat([Fragment_matrix, Amplicons], axis = 1, join = "outer")
        else:
            with get_context("spawn").Pool(processes = min([CPU, len(file_paths)]), initializer = init_worker) as pl:
                try:
                    Amplicons = list(tqdm(pl.imap(partial(add_anal_worker_process, fwd = fwd, rev = rev, amplicon_size_limit = amplicon_size_limit, overlap_region = overlap_region, circular_index = circular_index, Score_calculation = Score_calculation), file_paths), total = len(file_paths), desc = "Now analyzing ...", leave = False, unit = "tasks", unit_scale = True))
                except KeyboardInterrupt:
                    pl.terminate()
                    pl.join()
                    pl.close()
                    print("\n\n --- Keyboard Interrupt ---")
                    sys.exit()
            for amp in Amplicons:
                Fragment_matrix = pd.concat([Fragment_matrix, amp], axis = 1, join = "outer")
        if Dataset_type == 'Universal':
            Fragment_matrix = Fragment_matrix if np.any(Fragment_matrix.fillna('').applymap(lambda x:len(x) > 1)) else Fragment_matrix.applymap(lambda x:x[0] if type(x) is list else x)
        else:
            if Score_calculation == 'Sequence':
                for i in range(Fragment_matrix.shape[0]):
                    Fragment_matrix.iloc[i] = np.array(['' if np.all(pd.isna(Fragment_matrix.iloc[i, j])) else str(Fragment_matrix.iloc[i, j]) if calculate_score(Fragment_matrix.index[i][0], str(Fragment_matrix.iloc[i, j])[0:len(Fragment_matrix.index[i][0])]) == 1 else str(complementary_sequence(Fragment_matrix.iloc[i, j])) for j in range(len(Fragment_matrix.iloc[i]))], dtype = str)
            else:
                for i in range(Fragment_matrix.shape[0]):
                    Fragment_matrix.iloc[i] = np.array([str(sorted(list(set([(x, round(Fragment_matrix.iloc[i, j].count(x)/len(Fragment_matrix.iloc[i, j]), 2),) for x in Fragment_matrix.iloc[i, j]])))) if np.all(pd.notna(Fragment_matrix.iloc[i, j])) else '' for j in range(len(Fragment_matrix.iloc[i]))], dtype = object)
        if (csv_file_path != '') & (('max_col' in locals()) | (len(np.where(test_sequence.iloc[:, 0].isna() == True)[0].tolist()) == 1) | isSummary_data):
            Fragment_matrix = Fragment_matrix.add_prefix("[Additional_analysis]")
            if 'max_col' in locals():
                Template_matrix = pd.read_csv(csv_file_path, names = range(max_col), header = None, sep = ",")
            else:
                Template_matrix = pd.read_csv(csv_file_path, header = None, sep = ",")
            Template_matrix.index = pd.MultiIndex.from_tuples(zip(Template_matrix.iloc[:, 0], Template_matrix.iloc[:, 1]))
            Template_matrix = Template_matrix.drop(Template_matrix.columns[[0,1]], axis = 1)
            Template_analisys_information = Template_matrix[Template_matrix.isnull().all(axis = 1)]
            Template_matrix = Template_matrix[~Template_matrix.isnull().all(axis = 1)]
            Template_matrix.columns = Template_matrix.iloc[0]
            Template_matrix = Template_matrix.drop(Template_matrix.index[0])
            Template_matrix.index = Template_matrix.index.rename(Template_matrix.columns.names[0])
            Template_matrix.columns.name = None
            New_label = (['No'] if 'No' in Template_matrix.columns else []) + list(Fragment_matrix.columns) + list(Template_matrix.columns[Template_matrix.columns != 'No'] if 'No' in Template_matrix.columns else Template_matrix.columns)
            Fragment_matrix = pd.concat([Template_matrix, Fragment_matrix], axis = 1, join = "outer")
            Fragment_matrix = Fragment_matrix.reindex(columns = New_label)
            Fragment_matrix.index = Fragment_matrix.index.rename(['Forward', 'Reverse'])
            if (Dataset_type == 'Identify') & ("Score" in Fragment_matrix.columns):
                if Score_calculation == 'Sequence':
                    Reevaluation_matrix = Fragment_matrix.loc[:,Fragment_matrix.columns[Fragment_matrix.applymap(lambda x:len(re.sub("[ATGCBDHKMRSVWYNXatgcbdhkmrsvwynx]", "", str(x))) == 0 if np.all(pd.notna(x)) else False).all(axis = 0)]]
                    Reevaluation_matrix = Reevaluation_matrix.applymap(lambda x:str(x) if np.all(pd.notna(x)) else np.nan)
                    with get_context("spawn").Pool(processes = CPU, initializer = init_worker) as pl:
                        try:
                            Score = list(pl.imap(partial(cumulative_pairwise_identity, df = Reevaluation_matrix, method = 'average'), range(Reevaluation_matrix.shape[0])))
                            Score = [round(sc, 4) for sc in Score]
                        except KeyboardInterrupt:
                            pl.terminate()
                            pl.join()
                            pl.close()
                            print("\n\n --- Keyboard Interrupt ---")
                            sys.exit()
                    Fragment_matrix.insert(list(Fragment_matrix.columns).index("Score") + 1, "Recalculated_score", Score)
                else:
                    Reevaluation_matrix = Fragment_matrix.loc[:,Fragment_matrix.columns[Fragment_matrix.applymap(lambda x:True if x is None else True if np.all(pd.isna(x)) else np.any([x.count("[") > 0, x.lower() == 'nan', len(x) == 0]) if type(x) is str else type(x) is list).all(axis = 0)]]
                    Reevaluation_matrix = Reevaluation_matrix.applymap(lambda x:np.nan if np.any([x is None, x == '']) else x)
                    Reevaluation_matrix = Reevaluation_matrix.applymap(lambda x:eval(x) if (type(x) is str) else x)
                    if Reevaluation_matrix.applymap(lambda x:type(x[0]) is tuple if type(x) is list else False).any().any():
                        Reevaluation_matrix = Reevaluation_matrix.applymap(lambda x:[x1[0] for x1 in x] if np.all(pd.notna(x)) else np.nan)
                    Score = Reevaluation_matrix.apply(lambda x:round(fragment_size_distance(x[np.logical_not(x == '')], sum = False, method = 'average'), 1), axis = 1)
                    Fragment_number = Reevaluation_matrix.apply(lambda x:len(list(it.chain.from_iterable(x[pd.notna(x)]))), axis = 1)
                    Score = Score / (Fragment_number - np.min(Fragment_number) + 1)
                    Fragment_matrix.insert(list(Fragment_matrix.columns).index("Score") + 1, "Recalculated_score", Score)
                    Fragment_matrix.insert(list(Fragment_matrix.columns).index("Fragment_number") + 1, "Recalculated_fragment_number", Fragment_number)
        if output_folder != "":
            if not os.path.exists(output_folder):
                os.mkdir(output_folder)
            if (csv_file_path != '') & (('max_col' in locals()) | (len(np.where(test_sequence.iloc[:, 0].isna() == True)[0].tolist()) == 1) | isSummary_data):
                Template_analisys_information.to_csv(output_folder + csv_file_path[int(np.max([csv_file_path.rfind("/"), csv_file_path.rfind("\\")])) + 1 : csv_file_path.rfind(".csv")] + "_Additional_analysis_result.csv", header = False, sep = ",")
                Fragment_matrix.to_csv(output_folder + csv_file_path[int(np.max([csv_file_path.rfind("/"), csv_file_path.rfind("\\")])) + 1 : csv_file_path.rfind(".csv")] + "_Additional_analysis_result.csv", mode = "a", header = True, index = True, sep = ",")
            elif (csv_file_path != ''):
                Fragment_matrix.to_csv(output_folder + csv_file_path[int(np.max([csv_file_path.rfind("/"), csv_file_path.rfind("\\")])) + 1 : csv_file_path.rfind(".csv")] + "_Additional_analysis_result.csv", header = True, index = True, sep = ",")
            else:
                Fragment_matrix.to_csv(output_folder + "Additional_analysis_result.csv", sep = ",")
        elif csv_file_path == '':
            if os.path.isdir(input_file):
                if input_file.endswith("/"):
                    Fragment_matrix.to_csv(input_file+"Additional_analysis_result.csv", header = True, index = True, sep = ",")
                else:
                    Fragment_matrix.to_csv(input_file+"/Additional_analysis_result.csv", header = True, index = True, sep = ",")
            else:
                if input_file.find("/") == -1:
                    Fragment_matrix.to_csv("./Additional_analysis_result.csv", header = True, index = True, sep = ",")
                else:
                    Fragment_matrix.to_csv(input_file[0:input_file.rfind("/")]+"/Additional_analysis_result.csv", header = True, index = True, sep = ",")
        else:
            if (csv_file_path != '') & (('max_col' in locals()) | (len(np.where(test_sequence.iloc[:, 0].isna() == True)[0].tolist()) == 1) | isSummary_data):
                Template_analisys_information.to_csv(csv_file_path[0:csv_file_path.rfind(".csv")]+"_Additional_analysis_result.csv", header = False, sep = ",")
                Fragment_matrix.to_csv(csv_file_path[0:csv_file_path.rfind(".csv")]+"_Additional_analysis_result.csv", mode = "a", header = True, index = True, sep = ",")
            elif (csv_file_path != ''):
                Fragment_matrix.to_csv(csv_file_path[0:csv_file_path.rfind(".csv")]+"_Additional_analysis_result.csv", header = True, index = True, sep = ",")
            else:
                Fragment_matrix.to_csv(csv_file_path[0:csv_file_path.rfind(".csv")]+"_Additional_analysis_result.csv", header = True, index = True, sep = ",")
    else:
        print("Unexpected data format. Check CSV file.")
        sys.exit()


'''
Copyright (c) 2022 Masayuki TAKAHASHI

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), 
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

def insilicoPCR(args):
    """
    in silico PCR amplification algorithm.

    Parameters
    ----------
    args: ArgumentParser, class containing appropriate property as below
        input_file: Input file path (required)\n
        output: Output directory path (Make a new directory if the directory does not exist)\n
        size_limit: The upper limit of amplicon size\n
        process: The number of processes (sometimes the number of CPU core) used for analysis\n
        fasta: Output format. A FASTA file will be generated if you use this option.\n
        forward: The forward primer sequence used for amplification (required)\n
        reverse: The reverse primer sequence used for amplification (required)\n
        Single_file: Output format. One single FASTA-format file will be generated even if you input some separate FASTA files, when using this option with the '--fasta' option.\n
        Mismatch_allowance: The acceptable mismatch number\n
        Only_one_amplicon: Only one amplicon is outputted, even if multiple amplicons are obtained by PCR when you use this option.\n
        Position_index: The result has the information of the amplification position when this option is enabled.\n
        circularDNA: If there are some circular DNAs in the input sequences, use this option (default: n/a. It means all input sequences are linear DNA. When there are some circular DNA input sequences, type 'all', 'individually', 'n/a', or the file path of the text that specify which sequence is circularDNA, after the '--circularDNA' option. See Readme for more detailed information.)\n
            Text file example:\n
                Sequence_name1 circularDNA\n
                Sequence_name2 linearDNA\n
                Sequence_name3 linearDNA\n
                    ...\n
                Sequence_nameN circularDNA\n
        gene_annotation_search_range: The gene annotation search range in the GenBank-format file.\n
        Annotation: If the input sequence file is in GenBank format, the amplicon(s) is annotated automatically.\n
        warning: Shows all warnings when you use this option.\n

    Returns
    -------
    FASTA/CSV: Amplified sequence(s) FASTA format file (or CSV file)

    """
    import sys
    import os
    import re
    import warnings
    from functools import partial
    sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
    from tqdm import tqdm
    import math
    import itertools as it
    import numpy as np
    import pandas as pd
    from multiprocessing import Pool, get_context, cpu_count
    from shrslib.basicfunc import init_worker, read_sequence_file, circularDNA, check_input_file
    from shrslib.explore import PCR_amplicon
    from shrslib.multiprocessfunc import trim_seq_worker_process_all, trim_seq_worker_process_all_df, trim_seq_worker_process_single, trim_seq_worker_process_single_df, PCR_amplicon_with_progress_bar
    try:
        input_file = args.input_file
    except:
        sys.exit("Need a FASTA file or folder containing FASTA files after '-i '. \nPlease use '-h' or '--help' option if you would like to confirm usage.")
    if os.path.isfile(input_file):
        file_paths = [input_file]
    elif os.path.isdir(input_file):
        file_names = os.listdir(input_file)
        if input_file.endswith("/"):
            file_paths = [input_file+file_name for file_name in file_names]
        else:
            file_paths = [input_file+"/"+file_name for file_name in file_names]
        file_paths = [file_path for file_path in file_paths if os.path.isfile(file_path)]
        for file_path in file_paths:
            with open(file_path, "rt", encoding = "utf-8") as fin:
                Firstline = fin.readline()
                if np.logical_not(Firstline.startswith(">") | Firstline.startswith("LOCUS")):
                    print("There are some file(s) of which are not FASTA or Genbank in folder inputed. Please remove the file(s).")
                    sys.exit()
    else:
        sys.exit("Error: The input file/folder does not exist.")
    if args.output is not None:
        output_folder = args.output
        if not output_folder.endswith("/"):
            output_folder = output_folder+"/"
    else:
        output_folder = ""
    output_dataframe = args.fasta
    single_file = args.Single_file
    fwd = args.forward
    rev = args.reverse
    amplicon_size_limit = args.size_limit
    allowance = args.Mismatch_allowance
    Remain_all_amplicons = args.Only_one_amplicon
    Position_index = args.Position_index
    if args.Annotation:
        Feature = True
        Position_index = True
    else:
        Feature = False
    CPU = args.process
    if CPU is not None:
        if CPU > cpu_count():
            CPU = int(math.ceil(cpu_count()))
    else:
        if cpu_count() <= 3:
            CPU = int(1)
        elif cpu_count() <= 8:
            CPU = int(math.floor(cpu_count() / 2))
        elif cpu_count() <= 12:
            CPU = int(cpu_count() - 4)
        elif cpu_count() <= 24:
            CPU = int(math.floor(cpu_count() / 2) + 2)
        else:
            CPU = int(16)
    if args.circularDNA is not None:
        overlap_region = amplicon_size_limit
        if len(args.circularDNA) == 0:
            circular_index = check_input_file(file_paths, circular = "all")
        elif args.circularDNA[0].lower() == "all":
            circular_index = check_input_file(file_paths, circular = "all")
        elif args.circularDNA[0].lower() == "individually":
            circular_index = check_input_file(file_paths, circular = "individually")
        elif os.path.isfile(args.circularDNA[0]):
            circular_index = check_input_file(file_paths, circular = args.circularDNA[0])
        elif args.circularDNA[0].lower() == "n/a":
            overlap_region = 0
            circular_index = check_input_file(file_paths, circular = "n/a")
        else:
            print(" Specify the file path, 'all', 'individually' or 'n/a' after --circularDNA option.\n All input sequence(s) are analysed as circular DNA")
            circular_index = check_input_file(file_paths, circular = "all")
    else:
        overlap_region = 0
        circular_index = check_input_file(file_paths, circular = "n/a")
    distance = args.gene_annotation_search_range
    Warning_ignore = args.warning
    if Warning_ignore:
        warnings.simplefilter('ignore')
    Seq_no = []
    for file_path in file_paths:
        with open(file_path, "rt", encoding = "utf-8") as fin:
            read_FILE = fin.read()
            if read_FILE.startswith(">"):
                Seq_no += [read_FILE.count(">")]
            else:
                Seq_no += [1]
    annotation = {}
    if output_dataframe:
        Amplicon_Sequences = dict()
        if Remain_all_amplicons:
            if np.mean(Seq_no) > len(Seq_no):
                with tqdm(total = np.sum(Seq_no), desc = "Dataframe of amplified sequences making", leave = False, unit = "tasks", unit_scale = True) as pbar:
                    def Progressbar_update_alpha(Result):
                        pbar.update(Result[1])
                    for file_path in file_paths:
                        if Feature:
                            seq_dict = read_sequence_file(input = file_path, format = "genbank", Feature = True)[0]
                            annotation.update({key:seq_dict[key][1] for key in seq_dict})
                            seq_dict = {key:seq_dict[key][0] for key in seq_dict}
                        else:
                            seq_dict = read_sequence_file(file_path)[0]
                        Sequence_length_list = [np.min([amplicon_size_limit, seq_dict[key].sequence_length]) for key in seq_dict]
                        seq_dict = {key:circularDNA(sequence = seq_dict[key], overlap = int(np.min([overlap_region, seq_dict[key].sequence_length]))) if circular_index[key] else circularDNA(sequence = seq_dict[key], overlap = int(0)) for key in seq_dict}
                        if (CPU > 1):
                            with get_context("spawn").Pool(processes = min([CPU, len(seq_dict)]), initializer = init_worker) as pl:
                                try:
                                    if Position_index:
                                        Amplicon = [pl.apply_async(PCR_amplicon_with_progress_bar, kwds = {"forward": fwd, "reverse": rev, "template": seq, "Single_amplicon": False, "Sequence_Only": False, "amplicon_size_limit": limit, "allowance": allowance, "Warning_ignore": Warning_ignore}, callback = Progressbar_update_alpha) for limit, seq in zip(Sequence_length_list, seq_dict.values())]
                                    else:
                                        Amplicon = [pl.apply_async(PCR_amplicon_with_progress_bar, kwds = {"forward": fwd, "reverse": rev, "template": seq, "Single_amplicon": False, "Sequence_Only": True, "amplicon_size_limit": limit, "allowance": allowance, "Warning_ignore": Warning_ignore}, callback = Progressbar_update_alpha) for limit, seq in zip(Sequence_length_list, seq_dict.values())]
                                except KeyboardInterrupt:
                                    pl.terminate()
                                    pl.join()
                                    pl.close()
                                    print("\n\n --- Keyboard Interrupt ---")
                                    sys.exit("")
                                Amplicon = [amp.get()[0] for amp in Amplicon]
                        else:
                            if Position_index:
                                Amplicon = [(PCR_amplicon(forward = fwd, reverse = rev, template = seq, Single_amplicon = False, Sequence_Only = False, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore), pbar.update(1), ) for limit, seq in zip(Sequence_length_list, seq_dict.values())]
                            else:
                                Amplicon = [(PCR_amplicon(forward = fwd, reverse = rev, template = seq, Single_amplicon = False, Sequence_Only = True, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore), pbar.update(1), ) for limit, seq in zip(Sequence_length_list, seq_dict.values())]
                            Amplicon = [amp[0] for amp in Amplicon]
                        Amplicon = [[] if seq is None else seq for seq in Amplicon]
                        Amplicon = {name:amp for name, amp in zip(seq_dict.keys(), Amplicon)}
                        Amplicon_Sequences.update(Amplicon)
                    max_col = max([len(i) for i in Amplicon_Sequences.values()])
                    Amplicon_Sequences = {name:[seqs[i] if len(seqs) > i else np.nan for i in range(max_col)] for name, seqs in Amplicon_Sequences.items()}
                    Result = pd.DataFrame(Amplicon_Sequences).T
                    Result.columns = ["Amplicon_sequence_"+ str(i + 1) for i in range(max_col)]
            else:
                Result = dict()
                if Feature:
                    for file_path in file_paths:
                        seq_dict = read_sequence_file(input = file_path, format = "genbank", Feature = True)[0]
                        annotation.update({key:seq_dict[key][1] for key in seq_dict})
                        del seq_dict
                with get_context("spawn").Pool(processes = np.min([CPU, len(file_paths)]), initializer = init_worker) as pl:
                    try:
                        if Position_index:
                            Amplicons = list(tqdm(pl.imap(partial(trim_seq_worker_process_all_df, fwd = fwd, rev = rev, Sequence_Only = False, amplicon_size_limit = amplicon_size_limit, allowance = allowance, Warning_ignore = Warning_ignore, overlap_region = overlap_region, circular_index = circular_index), file_paths), total = np.sum(Seq_no), desc = "Dataframe of amplified sequences making", leave = False, unit = "tasks", unit_scale = True))
                        else:
                            Amplicons = list(tqdm(pl.imap(partial(trim_seq_worker_process_all_df, fwd = fwd, rev = rev, Sequence_Only = True, amplicon_size_limit = amplicon_size_limit, allowance = allowance, Warning_ignore = Warning_ignore, overlap_region = overlap_region, circular_index = circular_index), file_paths), total = np.sum(Seq_no), desc = "Dataframe of amplified sequences making", leave = False, unit = "tasks", unit_scale = True))
                    except KeyboardInterrupt:
                        pl.terminate()
                        pl.join()
                        pl.close()
                        print("\n\n --- Keyboard Interrupt ---")
                        sys.exit("")
                [Result.update(amplicon) for amplicon in Amplicons]
                Result = pd.DataFrame(Result).T
            if Position_index & Feature & (len(annotation) > 0):
                Amplicon_annotation = {}
                for i in range(len(Result.index)):
                    Gene = [[position for position in re.sub(r"[^0-9]+", " ", target).split(" ") if len(position) != 0] + [annotation[Result.index[i]][target]] for target in annotation[Result.index[i]].keys() if (Result.index[i] in annotation.keys())]
                    Amplicon_annotation.update({Result.index[i]:{"Annotation_of_amplicon_sequence_" + str(j + 1): ["{0}..{1}: {2}".format(g[0], g[1], g[2]) for g in Gene if (((int(g[0]) - int(distance)) <= np.abs(Result.iloc[i, j][0])) & ((int(g[1]) + int(distance)) >= np.abs(Result.iloc[i, j][0]))) | (((int(g[0]) - int(distance)) <= np.abs(Result.iloc[i, j][1])) & ((int(g[1]) + int(distance)) >= np.abs(Result.iloc[i, j][1]))) | (((int(g[0]) - int(distance)) >= np.abs(Result.iloc[i, j][0])) & ((int(g[1]) + int(distance)) <= np.abs(Result.iloc[i, j][1])))] if pd.notna(Result.iloc[i, j]) else np.nan for j in range(Result.shape[1])}})
                Amplicon_annotation = pd.DataFrame(Amplicon_annotation).T
                Result = pd.concat([Result, Amplicon_annotation],axis = 1)
        else:
            if np.mean(Seq_no) > len(Seq_no):
                with tqdm(total = np.sum(Seq_no), desc = "Dataframe of amplified sequences making", leave = False, unit = "tasks", unit_scale = True) as pbar:
                    def Progressbar_update_beta(Result):
                        pbar.update(Result[1])
                    for file_path in file_paths:
                        if Feature:
                            seq_dict = read_sequence_file(input = file_path, format = "genbank", Feature = True)[0]
                            annotation.update({key:seq_dict[key][1] for key in seq_dict})
                            seq_dict = {key:seq_dict[key][0] for key in seq_dict}
                        else:
                            seq_dict = read_sequence_file(file_path)[0]
                        Sequence_length_list = [np.min([amplicon_size_limit, seq_dict[key].sequence_length]) for key in seq_dict]
                        seq_dict = {key:circularDNA(sequence = seq_dict[key], overlap = int(np.min([overlap_region, seq_dict[key].sequence_length]))) if circular_index[key] else circularDNA(sequence = seq_dict[key], overlap = int(0)) for key in seq_dict}
                        if (CPU > 1):
                            with get_context("spawn").Pool(processes = min([CPU, len(seq_dict)]), initializer = init_worker) as pl:
                                try:
                                    if Position_index:
                                        Amplicon = [pl.apply_async(PCR_amplicon_with_progress_bar, kwds = {"forward": fwd, "reverse": rev, "template": seq, "Single_amplicon": True, "Sequence_Only": False, "amplicon_size_limit": limit, "allowance": allowance, "Warning_ignore": Warning_ignore}, callback = Progressbar_update_beta) for limit, seq in zip(Sequence_length_list, seq_dict.values())]
                                    else: 
                                        Amplicon = [pl.apply_async(PCR_amplicon_with_progress_bar, kwds = {"forward": fwd, "reverse": rev, "template": seq, "Single_amplicon": True, "Sequence_Only": True, "amplicon_size_limit": limit, "allowance": allowance, "Warning_ignore": Warning_ignore}, callback = Progressbar_update_beta) for limit, seq in zip(Sequence_length_list, seq_dict.values())]
                                except KeyboardInterrupt:
                                    pl.terminate()
                                    pl.join()
                                    pl.close()
                                    print("\n\n --- Keyboard Interrupt ---")
                                    sys.exit("")
                                Amplicon = [amp.get()[0] for amp in Amplicon]
                        else:
                            if Position_index:
                                Amplicon = [(PCR_amplicon(forward = fwd, reverse = rev, template = seq, Single_amplicon = True, Sequence_Only = False, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore), pbar.update(1), ) for limit, seq in zip(Sequence_length_list, seq_dict.values())]
                            else:
                                Amplicon = [(PCR_amplicon(forward = fwd, reverse = rev, template = seq, Single_amplicon = True, Sequence_Only = True, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore), pbar.update(1), ) for limit, seq in zip(Sequence_length_list, seq_dict.values())]
                            Amplicon = [amp[0] for amp in Amplicon]
                        Amplicon = ['' if seq is None else seq for seq in Amplicon]
                        if Position_index:
                            len_seqs = [len(seq[2]) if len(seq) > 1 else 0 for seq in Amplicon]
                        else:
                            len_seqs = [len(seq) for seq in Amplicon]
                        Amplicon = list(zip(Amplicon, len_seqs))
                        Amplicon = {name:amp for name, amp in zip(seq_dict.keys(), Amplicon)}
                        Amplicon_Sequences.update(Amplicon)
                    Result = pd.DataFrame(Amplicon_Sequences).T
                    Result.columns = ["Amplicon_sequence", "Length"]
            else:
                Result = dict()
                if Feature:
                    for file_path in file_paths:
                        seq_dict = read_sequence_file(input = file_path, format = "genbank", Feature = True)[0]
                        annotation.update({key:seq_dict[key][1] for key in seq_dict})
                        del seq_dict
                with get_context("spawn").Pool(processes = np.min([CPU, len(file_paths)]), initializer = init_worker) as pl:
                    try:
                        if Position_index:
                            Amplicons = list(tqdm(pl.imap(partial(trim_seq_worker_process_single_df, fwd = fwd, rev = rev, Sequence_Only = False, amplicon_size_limit = amplicon_size_limit, allowance = allowance, Warning_ignore = Warning_ignore, overlap_region = overlap_region, circular_index = circular_index), file_paths), total = np.sum(Seq_no), desc = "Dataframe of amplified sequences making", leave = False, unit = "tasks", unit_scale = True))
                        else:
                            Amplicons = list(tqdm(pl.imap(partial(trim_seq_worker_process_single_df, fwd = fwd, rev = rev, Sequence_Only = True, amplicon_size_limit = amplicon_size_limit, allowance = allowance, Warning_ignore = Warning_ignore, overlap_region = overlap_region, circular_index = circular_index), file_paths), total = np.sum(Seq_no), desc = "Dataframe of amplified sequences making", leave = False, unit = "tasks", unit_scale = True))
                    except KeyboardInterrupt:
                        pl.terminate()
                        pl.join()
                        pl.close()
                        print("\n\n --- Keyboard Interrupt ---")
                        sys.exit("")
                [Result.update(amplicon) for amplicon in Amplicons]
                Result = pd.DataFrame(Result).T
                Result.columns = ["Amplicon_sequence", "Length"]
            if Position_index & Feature & (len(annotation) > 0):
                Amplicon_annotation = {}
                for i in range(len(Result.index)):
                    Gene = [[position for position in re.sub(r"[^0-9]+", " ", target).split(" ") if len(position) != 0] + [annotation[Result.index[i]][target]] for target in annotation[Result.index[i]].keys() if (Result.index[i] in annotation.keys())]
                    Amplicon_annotation.update({Result.index[i]: ["{0}..{1}: {2}".format(g[0], g[1], g[2]) for g in Gene if (((int(g[0]) - int(distance)) <= np.abs(Result.loc[Result.index[i], 'Amplicon_sequence'][0])) & ((int(g[1]) + int(distance)) >= np.abs(Result.loc[Result.index[i], 'Amplicon_sequence'][0]))) | (((int(g[0]) - int(distance)) <= np.abs(Result.loc[Result.index[i], 'Amplicon_sequence'][1])) & ((int(g[1]) + int(distance)) >= np.abs(Result.loc[Result.index[i], 'Amplicon_sequence'][1]))) | (((int(g[0]) - int(distance)) >= np.abs(Result.loc[Result.index[i], 'Amplicon_sequence'][0])) & ((int(g[1]) + int(distance)) <= np.abs(Result.loc[Result.index[i], 'Amplicon_sequence'][1])))] if pd.notna(Result.loc[Result.index[i], 'Amplicon_sequence']) else np.nan})
                Amplicon_annotation = pd.Series(Amplicon_annotation)
                Result = Result.assign(Annotation = Amplicon_annotation)
        if output_folder != "":
            if not os.path.exists(output_folder):
                os.mkdir(output_folder)
            Result.to_csv(output_folder+"iPCR_result.csv")
        else:
            if os.path.isfile(input_file):
                if input_file.rfind(".") == -1:
                    Result.to_csv(input_file+"_iPCR_result.csv")
                else:
                    Result.to_csv(input_file[0:input_file.rfind("."):]+"_iPCR_result.csv")
            else:
                if input_file.endswith("/"):
                    Result.to_csv(input_file + "iPCR_result.csv")
                else:
                    Result.to_csv(input_file + "/" + "iPCR_result.csv")
    else:
        trimmed_seqs = []
        if Remain_all_amplicons:
            if np.mean(Seq_no) > len(Seq_no):
                with tqdm(total = np.sum(Seq_no), desc = "Writing PCR result", leave = False, unit = "tasks", unit_scale = True) as pbar:
                    def Progressbar_update_gamma(Result):
                        pbar.update(Result[1])
                    for file_path in file_paths:
                        if Feature:
                            seq_dict = read_sequence_file(input = file_path, format = "genbank", Feature = True)[0]
                            annotation.update({key:seq_dict[key][1] for key in seq_dict})
                            seq_dict = {key:seq_dict[key][0] for key in seq_dict}
                        else:
                            seq_dict = read_sequence_file(file_path)[0]
                        Sequence_length_list = [np.min([amplicon_size_limit, seq_dict[key].sequence_length]) for key in seq_dict]
                        seq_dict = {key:circularDNA(sequence = seq_dict[key], overlap = int(np.min([overlap_region, seq_dict[key].sequence_length]))) if circular_index[key] else circularDNA(sequence = seq_dict[key], overlap = int(0)) for key in seq_dict}
                        if (CPU > 1):
                            with get_context("spawn").Pool(processes = np.min([CPU, len(seq_dict)]), initializer = init_worker) as pl:
                                try:
                                    if Position_index:
                                        trimmed_seq = [(file_path, name, pl.apply_async(PCR_amplicon_with_progress_bar, kwds = {"forward": fwd, "reverse": rev, "template": seq_dict[name], "Single_amplicon": False, "Sequence_Only": False, "amplicon_size_limit": limit, "allowance": allowance, "Warning_ignore": Warning_ignore}, callback = Progressbar_update_gamma), ) for limit, name in zip(Sequence_length_list, seq_dict)]
                                    else:
                                        trimmed_seq = [(file_path, name, pl.apply_async(PCR_amplicon_with_progress_bar, kwds = {"forward": fwd, "reverse": rev, "template": seq_dict[name], "Single_amplicon": False, "Sequence_Only": True, "amplicon_size_limit": limit, "allowance": allowance, "Warning_ignore": Warning_ignore}, callback = Progressbar_update_gamma), ) for limit, name in zip(Sequence_length_list, seq_dict)]
                                except KeyboardInterrupt:
                                    pl.terminate()
                                    pl.join()
                                    pl.close()
                                    print("\n\n --- Keyboard Interrupt ---")
                                    sys.exit("")
                                trimmed_seq = [(ts[0], ts[1], ts[2].get()[0], ) for ts in trimmed_seq]
                        else:
                            if Position_index:
                                trimmed_seq = [(file_path, name, PCR_amplicon(forward = fwd, reverse = rev, template = seq_dict[name], Single_amplicon = False, Sequence_Only = False, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore), pbar.update(1), ) for limit, name in zip(Sequence_length_list, seq_dict)]
                            else:
                                trimmed_seq = [(file_path, name, PCR_amplicon(forward = fwd, reverse = rev, template = seq_dict[name], Single_amplicon = False, Sequence_Only = True, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore), pbar.update(1), ) for limit, name in zip(Sequence_length_list, seq_dict)]
                        trimmed_seq = [ts for ts in trimmed_seq if ts[2] is not None]
                        if Position_index & Feature & (len(annotation) > 0): 
                            annotated_trimmed_seq = []
                            for ts in trimmed_seq: 
                                Gene = [[position for position in re.sub(r"[^0-9]+", " ", target).split(" ") if len(position) != 0] + [annotation[ts[1]][target]] for target in annotation[ts[1]].keys() if (ts[1] in annotation.keys())] 
                                Amplicon_annotation = {str(amp[0]) + str(amp[1]): [(g[0], g[1], g[2], ) for g in Gene if (((int(g[0]) - int(distance)) <= np.abs(amp[0])) & ((int(g[1]) + int(distance)) >= np.abs(amp[0]))) | (((int(g[0]) - int(distance)) <= np.abs(amp[1])) & ((int(g[1]) + int(distance)) >= np.abs(amp[1]))) | (((int(g[0]) - int(distance)) >= np.abs(amp[0])) & ((int(g[1]) + int(distance)) <= np.abs(amp[1])))] for amp in ts[2]} 
                                annotated_trimmed_seq.append((ts[0], ts[1], [tuple(list(amp) + [Amplicon_annotation[str(amp[0]) + str(amp[1])]]) for amp in ts[2]], )) 
                            trimmed_seq = ([[(ts[0], ts[1]+"_amplicon_"+str(j+1)+"_("+str(seq[0])+" -> "+str(seq[1])+")", seq[2], seq[3], ) for j, seq in enumerate(ts[2])] if len(ts[2]) > 0 else ts for ts in annotated_trimmed_seq]) if len(annotated_trimmed_seq) > 0 else annotated_trimmed_seq 
                            del annotated_trimmed_seq
                        elif Position_index:
                            trimmed_seq = ([[(ts[0], ts[1]+"_amplicon_"+str(j+1)+"_("+str(seq[0])+" -> "+str(seq[1])+")", seq[2], [], ) for j, seq in enumerate(ts[2])] if len(ts[2]) > 0 else ts for ts in trimmed_seq]) if len(trimmed_seq) > 0 else trimmed_seq
                        else:
                            trimmed_seq = ([[(ts[0], ts[1]+"_amplicon_"+str(j+1), seq, [], ) for j, seq in enumerate(ts[2])] if len(ts[2]) > 0 else ts for ts in trimmed_seq]) if len(trimmed_seq) > 0 else trimmed_seq
                        trimmed_seq = list(it.chain.from_iterable(trimmed_seq))
                        trimmed_seqs += trimmed_seq
            else:
                if Feature: 
                    for file_path in file_paths: 
                        seq_dict = read_sequence_file(input = file_path, format = "genbank", Feature = True)[0] 
                        annotation.update({key:seq_dict[key][1] for key in seq_dict}) 
                        del seq_dict 
                with get_context("spawn").Pool(processes = np.min([CPU, len(file_paths)]), initializer = init_worker) as pl:
                    try:
                        if Position_index & Feature & (len(annotation) > 0):
                            trimmed_seq = list(tqdm(pl.imap(partial(trim_seq_worker_process_all, fwd = fwd, rev = rev, Sequence_Only = False, amplicon_size_limit = amplicon_size_limit, allowance = allowance, Warning_ignore = Warning_ignore, overlap_region = overlap_region, circular_index = circular_index, Feature = Feature, annotation = annotation, distance = distance), file_paths), total = np.sum(Seq_no), desc = "Writing PCR result", leave = False, unit = "tasks", unit_scale = True))
                        elif Position_index:
                            trimmed_seq = list(tqdm(pl.imap(partial(trim_seq_worker_process_all, fwd = fwd, rev = rev, Sequence_Only = False, amplicon_size_limit = amplicon_size_limit, allowance = allowance, Warning_ignore = Warning_ignore, overlap_region = overlap_region, circular_index = circular_index), file_paths), total = np.sum(Seq_no), desc = "Writing PCR result", leave = False, unit = "tasks", unit_scale = True))
                        else:
                            trimmed_seq = list(tqdm(pl.imap(partial(trim_seq_worker_process_all, fwd = fwd, rev = rev, Sequence_Only = True, amplicon_size_limit = amplicon_size_limit, allowance = allowance, Warning_ignore = Warning_ignore, overlap_region = overlap_region, circular_index = circular_index), file_paths), total = np.sum(Seq_no), desc = "Writing PCR result", leave = False, unit = "tasks", unit_scale = True))
                    except KeyboardInterrupt:
                        pl.terminate()
                        pl.join()
                        pl.close()
                        print("\n\n --- Keyboard Interrupt ---")
                        sys.exit("")
                trimmed_seqs = list(it.chain.from_iterable(trimmed_seq))
        else:
            if np.mean(Seq_no) > len(Seq_no):
                with tqdm(total = np.sum(Seq_no), desc = "Writing PCR result", leave = False, unit = "tasks", unit_scale = True) as pbar:
                    def Progressbar_update_delta(Result):
                        pbar.update(Result[1])
                    for file_path in file_paths:
                        if Feature:
                            seq_dict = read_sequence_file(input = file_path, format = "genbank", Feature = True)[0]
                            annotation.update({key:seq_dict[key][1] for key in seq_dict})
                            seq_dict = {key:seq_dict[key][0] for key in seq_dict}
                        else:
                            seq_dict = read_sequence_file(file_path)[0]
                        Sequence_length_list = [np.min([amplicon_size_limit, seq_dict[key].sequence_length]) for key in seq_dict]
                        seq_dict = {key:circularDNA(sequence = seq_dict[key], overlap = int(np.min([overlap_region, seq_dict[key].sequence_length]))) if circular_index[key] else circularDNA(sequence = seq_dict[key], overlap = int(0)) for key in seq_dict}
                        if (CPU > 1):
                            with get_context("spawn").Pool(processes = np.min([CPU, len(seq_dict)]), initializer = init_worker) as pl:
                                try:
                                    if Position_index:
                                        trimmed_seq = [(file_path, name, pl.apply_async(PCR_amplicon_with_progress_bar, kwds = {"forward": fwd, "reverse": rev, "template": seq_dict[name], "Single_amplicon": True, "Sequence_Only": False, "amplicon_size_limit": limit, "allowance": allowance, "Warning_ignore": Warning_ignore}, callback = Progressbar_update_delta), ) for limit, name in zip(Sequence_length_list, seq_dict)]
                                    else:
                                        trimmed_seq = [(file_path, name, pl.apply_async(PCR_amplicon_with_progress_bar, kwds = {"forward": fwd, "reverse": rev, "template": seq_dict[name], "Single_amplicon": True, "Sequence_Only": True, "amplicon_size_limit": limit, "allowance": allowance, "Warning_ignore": Warning_ignore}, callback = Progressbar_update_delta), ) for limit, name in zip(Sequence_length_list, seq_dict)]
                                except KeyboardInterrupt:
                                    pl.terminate()
                                    pl.join()
                                    pl.close()
                                    print("\n\n --- Keyboard Interrupt ---")
                                    sys.exit("")
                                trimmed_seq = [(ts[0], ts[1], ts[2].get()[0], ) for ts in trimmed_seq]
                        else:
                            if Position_index:
                                trimmed_seq = [(file_path, name, PCR_amplicon(forward = fwd, reverse = rev, template = seq_dict[name], Single_amplicon = True, Sequence_Only = False, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore), pbar.update(1), ) for limit, name in zip(Sequence_length_list, seq_dict)]
                            else:
                                trimmed_seq = [(file_path, name, PCR_amplicon(forward = fwd, reverse = rev, template = seq_dict[name], Single_amplicon = True, Sequence_Only = True, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore), pbar.update(1), ) for limit, name in zip(Sequence_length_list, seq_dict)]
                        trimmed_seq = [ts for ts in trimmed_seq if ts[2] is not None]
                        if Position_index & Feature & (len(annotation) > 0): 
                            annotated_trimmed_seq = []
                            for ts in trimmed_seq: 
                                Gene = [[position for position in re.sub(r"[^0-9]+", " ", target).split(" ") if len(position) != 0] + [annotation[ts[1]][target]] for target in annotation[ts[1]].keys() if (ts[1] in annotation.keys())] 
                                Amplicon_annotation = {str(ts[2][0]) + str(ts[2][1]): [(g[0], g[1], g[2], ) for g in Gene if (((int(g[0]) - int(distance)) <= np.abs(ts[2][0])) & ((int(g[1]) + int(distance)) >= np.abs(ts[2][0]))) | (((int(g[0]) - int(distance)) <= np.abs(ts[2][1])) & ((int(g[1]) + int(distance)) >= np.abs(ts[2][1]))) | (((int(g[0]) - int(distance)) >= np.abs(ts[2][0])) & ((int(g[1]) + int(distance)) <= np.abs(ts[2][1])))]} 
                                annotated_trimmed_seq.append((ts[0], ts[1], tuple(list(ts[2]) + [Amplicon_annotation[str(ts[2][0]) + str(ts[2][1])]]), )) 
                            trimmed_seq = [(ts[0], ts[1]+"_("+str(ts[2][0])+" -> "+str(ts[2][1])+")", ts[2][2], ts[2][3], ) for ts in annotated_trimmed_seq] 
                            del annotated_trimmed_seq
                        elif Position_index:
                            trimmed_seq = [(ts[0], ts[1]+"_("+str(ts[2][0])+" -> "+str(ts[2][1])+")", ts[2][2], [], ) for ts in trimmed_seq]
                        else:
                            trimmed_seq = [(ts[0], ts[1], ts[2], [], ) for ts in trimmed_seq]
                        trimmed_seqs += trimmed_seq
            else:
                if Feature: 
                    for file_path in file_paths: 
                        seq_dict = read_sequence_file(input = file_path, format = "genbank", Feature = True)[0] 
                        annotation.update({key:seq_dict[key][1] for key in seq_dict}) 
                        del seq_dict 
                with get_context("spawn").Pool(processes = np.min([CPU, len(file_paths)]), initializer = init_worker) as pl:
                    try:
                        if Position_index & Feature & (len(annotation) > 0):
                            trimmed_seqs = list(tqdm(pl.imap(partial(trim_seq_worker_process_single, fwd = fwd, rev = rev, Sequence_Only = False, amplicon_size_limit = amplicon_size_limit, allowance = allowance, Warning_ignore = Warning_ignore, overlap_region = overlap_region, circular_index = circular_index, Feature = Feature, annotation = annotation, distance = distance), file_paths), total = np.sum(Seq_no), desc = "Writing PCR result", leave = False, unit = "tasks", unit_scale = True))
                        elif Position_index:
                            trimmed_seqs = list(tqdm(pl.imap(partial(trim_seq_worker_process_single, fwd = fwd, rev = rev, Sequence_Only = False, amplicon_size_limit = amplicon_size_limit, allowance = allowance, Warning_ignore = Warning_ignore, overlap_region = overlap_region, circular_index = circular_index), file_paths), total = np.sum(Seq_no), desc = "Writing PCR result", leave = False, unit = "tasks", unit_scale = True))
                        else:
                            trimmed_seqs = list(tqdm(pl.imap(partial(trim_seq_worker_process_single, fwd = fwd, rev = rev, Sequence_Only = True, amplicon_size_limit = amplicon_size_limit, allowance = allowance, Warning_ignore = Warning_ignore, overlap_region = overlap_region, circular_index = circular_index), file_paths), total = np.sum(Seq_no), desc = "Writing PCR result", leave = False, unit = "tasks", unit_scale = True))
                    except KeyboardInterrupt:
                        pl.terminate()
                        pl.join()
                        pl.close()
                        print("\n\n --- Keyboard Interrupt ---")
                        sys.exit("")
                trimmed_seqs = list(it.chain.from_iterable(trimmed_seqs))
        if len(trimmed_seqs) == 0:
            print("No amplicon was obtained.")
            sys.exit()
        if output_folder != "":
            if not os.path.exists(output_folder):
                os.mkdir(output_folder)
            if single_file:
                for trimmed_seq in trimmed_seqs:
                    with open(output_folder+"iPCR_result.fasta", "a", encoding = "utf-8") as fout:
                        fout.write(">"+trimmed_seq[1]+"\n"+trimmed_seq[2]+"\n")
            elif Feature:
                for trimmed_seq in trimmed_seqs:
                    with open(output_folder + re.sub(r'_(.)_', r'\1', re.sub(r' (.) ', r'\1', re.sub(r'\.|,|-|>', "", str(trimmed_seq[1])))).replace("  ", " ").replace(" ", "_") + ".gb", "wt", encoding = "utf-8") as fout:
                        seq =  str(trimmed_seq[2])
                        Gene = trimmed_seq[3]
                        split_sequence_list = [seq[i * 10:(i+1) * 10:] for i in range(len(seq) // 10 + 1)]
                        Sequence = "".join([" ".join([str(i * 60 + 1).rjust(9)] + split_sequence_list[i * 6 : (i + 1) * 6:] + ["\n"]) for i in range(sum([len(split_sequence) == 10 for split_sequence in split_sequence_list]) // 6 + 1)])
                        LOCUS = "LOCUS".ljust(12) + str(trimmed_seq[1])
                        DEFINITION = "DEFINITION".ljust(12) + str(trimmed_seq[1])
                        TITLE = "TITLE".ljust(12) + str(trimmed_seq[1])
                        FEATURES = "FEATURES             Location/Qualifiers"
                        annotation_template1 = "     CDS             {0}..{1}" + "\n" + "".ljust(21) + '/product="{2}"'
                        annotation_template2 = "     CDS             complement({1}..{0})" + "\n" + "".ljust(21) + '/product="{2}"'
                        ANNOTATION = "\n".join([annotation_template1.format(g[0], g[1], g[2]) if int(g[0]) > 0 else annotation_template2.format(g[0], g[1], g[2]) for g in Gene])
                        SEQUENCE = "ORIGIN\n\n" + Sequence + "//"
                        OUTPUT_DATA = LOCUS + "\n" + DEFINITION + "\n" + TITLE + "\n" + FEATURES + "\n" + ANNOTATION + "\n" + SEQUENCE
                        fout.write(OUTPUT_DATA)
            else:
                for trimmed_seq in trimmed_seqs:
                    with open(output_folder+trimmed_seq[0][trimmed_seq[0].rfind("/") + 1:].replace(trimmed_seq[0][trimmed_seq[0].rfind("."):], "") + "_iPCR_result.fasta", "a", encoding = "utf-8") as fout:
                        fout.write(">"+trimmed_seq[1]+"\n"+trimmed_seq[2]+"\n")
        else:
            if single_file:
                for trimmed_seq in trimmed_seqs:
                    with open(trimmed_seq[0].replace(trimmed_seq[0][trimmed_seq[0].rfind("/") + 1:], "") + "iPCR_result.fasta", "a", encoding = "utf-8") as fout:
                        fout.write(">"+trimmed_seq[1]+"\n"+trimmed_seq[2]+"\n")
            elif Feature:
                for trimmed_seq in trimmed_seqs:
                    with open(trimmed_seq[0].replace(trimmed_seq[0][trimmed_seq[0].rfind("/") + 1:], "") + re.sub(r'_(.)_', r'\1', re.sub(r' (.) ', r'\1', re.sub(r'\.|,|-|>', "", str(trimmed_seq[1])))).replace("  ", " ").replace(" ", "_") + ".gb", "wt", encoding = "utf-8") as fout:
                        seq =  str(trimmed_seq[2])
                        Gene = trimmed_seq[3]
                        split_sequence_list = [seq[i * 10:(i+1) * 10:] for i in range(len(seq) // 10 + 1)]
                        Sequence = "".join([" ".join([str(i * 60 + 1).rjust(9)] + split_sequence_list[i * 6 : (i + 1) * 6:] + ["\n"]) for i in range(sum([len(split_sequence) == 10 for split_sequence in split_sequence_list]) // 6 + 1)])
                        LOCUS = "LOCUS".ljust(12) + str(trimmed_seq[1])
                        DEFINITION = "DEFINITION".ljust(12) + str(trimmed_seq[1])
                        TITLE = "TITLE".ljust(12) + str(trimmed_seq[1])
                        FEATURES = "FEATURES             Location/Qualifiers"
                        annotation_template1 = "     CDS             {0}..{1}" + "\n" + "".ljust(21) + '/product="{2}"'
                        annotation_template2 = "     CDS             complement({1}..{0})" + "\n" + "".ljust(21) + '/product="{2}"'
                        ANNOTATION = "\n".join([annotation_template1.format(g[0], g[1], g[2]) if int(g[0]) > 0 else annotation_template2.format(g[0], g[1], g[2]) for g in Gene])
                        SEQUENCE = "ORIGIN\n\n" + Sequence + "//"
                        OUTPUT_DATA = LOCUS + "\n" + DEFINITION + "\n" + TITLE + "\n" + FEATURES + "\n" + ANNOTATION + "\n" + SEQUENCE
                        fout.write(OUTPUT_DATA)
            else:
                for trimmed_seq in trimmed_seqs:
                    with open(trimmed_seq[0].replace(trimmed_seq[0][trimmed_seq[0].rfind("."):], "") + "_iPCR_result.fasta", "a", encoding = "utf-8") as fout:
                        fout.write(">"+trimmed_seq[1]+"\n"+trimmed_seq[2]+"\n")


'''
Copyright (c) 2022 Masayuki TAKAHASHI

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), 
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

import os
import sys
import re
import copy
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
import numpy as np
import pandas as pd
import itertools as it
from shrslib.basicfunc import read_sequence_file, circularDNA, check_input_file
from shrslib.explore import PCR_amplicon
from shrslib.scores import pairwise_identity
from Bio import pairwise2
try:
    import cupy as cp
    mempool = cp.get_default_memory_pool()
except:
    pass

def trim_seq_worker_process_all(fp, fwd, rev, Sequence_Only, amplicon_size_limit ,allowance, Warning_ignore, overlap_region, circular_index, Feature = False, annotation = {}, distance = int(0)):
    """
    Function for generating fragment based on template DNA, forward and reverse primer set by parallel process. (private)
    
    """
    seq_dict = read_sequence_file(fp)[0]
    Sequence_length_list = [np.min([amplicon_size_limit, seq_dict[key].sequence_length]) for key in seq_dict]
    seq_dict = {key:circularDNA(sequence = seq_dict[key], overlap = int(np.min([overlap_region, seq_dict[key].sequence_length]))) if circular_index[key] else circularDNA(sequence = seq_dict[key], overlap = int(0)) for key in seq_dict}
    trimmed_seq = [(fp, name, PCR_amplicon(forward = fwd, reverse = rev, template = seq_dict[name], Single_amplicon = False, Sequence_Only = Sequence_Only, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore), ) for limit, name in zip(Sequence_length_list, seq_dict)]
    trimmed_seq = [ts for ts in trimmed_seq if ts[2] is not None]
    if np.logical_not(Sequence_Only) & Feature & (len(annotation) > 0): 
        annotated_trimmed_seq = []
        for ts in trimmed_seq: 
            Gene = [[position for position in re.sub(r"[^0-9]+", " ", target).split(" ") if len(position) != 0] + [annotation[ts[1]][target]] for target in annotation[ts[1]].keys() if (ts[1] in annotation.keys())] 
            Amplicon_annotation = {str(amp[0]) + str(amp[1]): [(g[0], g[1], g[2], ) for g in Gene if (((int(g[0]) - int(distance)) <= np.abs(amp[0])) & ((int(g[1]) + int(distance)) >= np.abs(amp[0]))) | (((int(g[0]) - int(distance)) <= np.abs(amp[1])) & ((int(g[1]) + int(distance)) >= np.abs(amp[1]))) | (((int(g[0]) - int(distance)) >= np.abs(amp[0])) & ((int(g[1]) + int(distance)) <= np.abs(amp[1])))] for amp in ts[2]} 
            annotated_trimmed_seq.append((ts[0], ts[1], [tuple(list(amp) + [Amplicon_annotation[str(amp[0]) + str(amp[1])]]) for amp in ts[2]], )) 
        trimmed_seq = ([[(ts[0], ts[1]+"_amplicon_"+str(j+1)+"_("+str(seq[0])+" -> "+str(seq[1])+")", seq[2], seq[3], ) for j, seq in enumerate(ts[2])] if len(ts[2]) > 0 else ts for ts in annotated_trimmed_seq]) if len(annotated_trimmed_seq) > 0 else annotated_trimmed_seq
        del annotated_trimmed_seq
    elif np.logical_not(Sequence_Only):
        trimmed_seq = ([[(ts[0], ts[1]+"_amplicon_"+str(j+1)+"_("+str(seq[0])+" -> "+str(seq[1])+")", seq[2], [], ) for j, seq in enumerate(ts[2])] if len(ts[2]) > 0 else ts for ts in trimmed_seq]) if len(trimmed_seq) > 0 else trimmed_seq
    else:
        trimmed_seq = ([[(ts[0], ts[1]+"_amplicon_"+str(j+1), seq, [], ) for j, seq in enumerate(ts[2])] if len(ts[2]) > 0 else ts for ts in trimmed_seq]) if len(trimmed_seq) > 0 else trimmed_seq
    trimmed_seq = list(it.chain.from_iterable(trimmed_seq))
    return trimmed_seq

def trim_seq_worker_process_single(fp, fwd, rev, Sequence_Only, amplicon_size_limit ,allowance, Warning_ignore, overlap_region, circular_index, Feature = False, annotation = {}, distance = int(0)):
    """
    Function for generating fragment based on template DNA, forward and reverse primer set by parallel process. (private)
    
    """
    seq_dict = read_sequence_file(fp)[0]
    Sequence_length_list = [np.min([amplicon_size_limit, seq_dict[key].sequence_length]) for key in seq_dict]
    seq_dict = {key:circularDNA(sequence = seq_dict[key], overlap = int(np.min([overlap_region, seq_dict[key].sequence_length]))) if circular_index[key] else circularDNA(sequence = seq_dict[key], overlap = int(0)) for key in seq_dict}
    trimmed_seq = [(fp, name, PCR_amplicon(forward = fwd, reverse = rev, template = seq_dict[name], Single_amplicon = True, Sequence_Only = Sequence_Only, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore), ) for limit, name in zip(Sequence_length_list, seq_dict)]
    trimmed_seq = [ts for ts in trimmed_seq if ts[2] is not None]
    if np.logical_not(Sequence_Only) & Feature & (len(annotation) > 0): 
        annotated_trimmed_seq = []
        for ts in trimmed_seq: 
            Gene = [[position for position in re.sub(r"[^0-9]+", " ", target).split(" ") if len(position) != 0] + [annotation[ts[1]][target]] for target in annotation[ts[1]].keys()] 
            Amplicon_annotation = {str(ts[2][0]) + str(ts[2][1]): [(g[0], g[1], g[2], ) for g in Gene if (((int(g[0]) - int(distance)) <= np.abs(ts[2][0])) & ((int(g[1]) + int(distance)) >= np.abs(ts[2][0]))) | (((int(g[0]) - int(distance)) <= np.abs(ts[2][1])) & ((int(g[1]) + int(distance)) >= np.abs(ts[2][1]))) | (((int(g[0]) - int(distance)) >= np.abs(ts[2][0])) & ((int(g[1]) + int(distance)) <= np.abs(ts[2][1])))]} 
            annotated_trimmed_seq.append((ts[0], ts[1], tuple(list(ts[2]) + [Amplicon_annotation[str(ts[2][0]) + str(ts[2][1])]]), ))
        trimmed_seq = [(ts[0], ts[1]+"_("+str(ts[2][0])+" -> "+str(ts[2][1])+")", ts[2][2], ts[2][3], ) for ts in annotated_trimmed_seq]
        del annotated_trimmed_seq
    elif np.logical_not(Sequence_Only):
        trimmed_seq = [(ts[0], ts[1]+"_("+str(ts[2][0])+" -> "+str(ts[2][1])+")", ts[2][2], [], ) for ts in trimmed_seq]
    else:
        trimmed_seq = [(ts[0], ts[1], ts[2], [], ) for ts in trimmed_seq]
    return trimmed_seq

def trim_seq_worker_process_all_df(fp, fwd, rev, Sequence_Only, amplicon_size_limit, allowance, Warning_ignore, overlap_region, circular_index):
    """
    Function for generating fragment based on template DNA, forward and reverse primer set by parallel process. (private)
    
    """
    seq_dict = read_sequence_file(fp)[0]
    Sequence_length_list = [np.min([amplicon_size_limit, seq_dict[key].sequence_length]) for key in seq_dict]
    seq_dict = {key:circularDNA(sequence = seq_dict[key], overlap = int(np.min([overlap_region, seq_dict[key].sequence_length]))) if circular_index[key] else circularDNA(sequence = seq_dict[key], overlap = int(0)) for key in seq_dict}
    Amplicon = [PCR_amplicon(forward = fwd, reverse = rev, template = seq, Single_amplicon = False, Sequence_Only = Sequence_Only, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore) for limit, seq in zip(Sequence_length_list, seq_dict.values())]
    Amplicon = [[] if seq is None else seq for seq in Amplicon]
    Amplicon = {name:{"Amplicon_sequence_"+str(i + 1):amp for i, amp in enumerate(amps)} if len(amps) > 0 else {"Amplicon_sequence_1":None} for name, amps in zip(seq_dict.keys(), Amplicon)}
    return Amplicon

def trim_seq_worker_process_single_df(fp, fwd, rev, Sequence_Only, amplicon_size_limit ,allowance, Warning_ignore, overlap_region, circular_index):
    """
    Function for generating fragment based on template DNA, forward and reverse primer set by parallel process. (private)
    
    """
    seq_dict = read_sequence_file(fp)[0]
    Sequence_length_list = [np.min([amplicon_size_limit, seq_dict[key].sequence_length]) for key in seq_dict]
    seq_dict = {key:circularDNA(sequence = seq_dict[key], overlap = int(np.min([overlap_region, seq_dict[key].sequence_length]))) if circular_index[key] else circularDNA(sequence = seq_dict[key], overlap = int(0)) for key in seq_dict}
    Amplicon = [PCR_amplicon(forward = fwd, reverse = rev, template = seq, Single_amplicon = True, Sequence_Only = Sequence_Only, amplicon_size_limit = limit, allowance = allowance, Warning_ignore = Warning_ignore) for limit, seq in zip(Sequence_length_list, seq_dict.values())]
    Amplicon = ['' if seq is None else seq for seq in Amplicon]
    Amplicon = {name:[amp, len(amp[2])] if type(amp) is tuple else [amp, len(amp)] for name, amp in zip(seq_dict.keys(), Amplicon)}
    return Amplicon

def add_anal_worker_process(fp, fwd, rev, amplicon_size_limit, overlap_region, circular_index, Score_calculation):
    """
    Function for generating fragment based on template DNA, forward and reverse primer set by parallel process. (private)
    
    """
    seq_dict = read_sequence_file(fp)[0]
    Sequence_length_list = [np.min([amplicon_size_limit, seq_dict[key].sequence_length]) for key in seq_dict]
    seq_dict = {key:circularDNA(sequence = seq_dict[key], overlap = int(np.min([overlap_region, seq_dict[key].sequence_length]))) if circular_index[key] else circularDNA(sequence = seq_dict[key], overlap = int(0)) for key in seq_dict}
    Result = pd.DataFrame([])
    for limit, name in zip(Sequence_length_list, seq_dict):
        Amplicons = [PCR_amplicon(forward = f, reverse = r, template = seq_dict[name], Single_amplicon = False, Sequence_Only = True, amplicon_size_limit = limit, allowance = int(0), Warning_ignore = True) for f, r in zip(fwd, rev)]
        if Score_calculation == 'Sequence':
            Amplicons = [amps if amps is not None else [] for amps in Amplicons]
            Amplicons = [amps[0] if len(amps) == 1 else np.nan for amps in Amplicons]
        else:
            Amplicons = [[len(amp) for amp in amps] if amps is not None else np.nan for amps in Amplicons]
        Amplicons = pd.DataFrame(pd.Series(Amplicons, index = pd.MultiIndex.from_tuples([mid for mid in zip(fwd, rev)]), name = name))
        Result = pd.concat([Result, Amplicons], axis = 1, join = "outer")
    return Result

def PCR_amplicon_with_progress_bar(*args, **kwargs):
    """
    Wrapper for using tqdm in PCR_amplicon function with parallel process. (private)
    
    """
    amplicon = PCR_amplicon(*args, **kwargs)
    return (amplicon, 1, )

def cumulative_pairwise_identity(i, df, method = 'average'):
    """
    Function for calculating pairwise identity by parallel process. (private)
    
    """
    try:
        Align_score_list = [pairwise_identity(pairwise2.align.globalxx(p1, p2)) for p1, p2 in it.combinations(df.iloc[i][np.logical_not(df.iloc[i] == '')], 2) if np.all(pd.notna([p1, p2]))]
        Align_sequence_number = [0 for p1, p2 in it.combinations(df.iloc[i][np.logical_not(df.iloc[i] == '')], 2) if np.all(pd.notna([p1, p2]))]
        Average_identity_score = np.exp(np.sum(np.log(Align_score_list)) / len(Align_sequence_number))
        Maximum_identity_score = np.max(Align_score_list)
        Minimum_identity_score = np.min(Align_score_list)
        if str(method).lower() == 'average':
            Identity_score = Average_identity_score
        elif str(method).lower() == 'max':
            Identity_score = Maximum_identity_score
        elif str(method).lower() == 'min':
            Identity_score = Minimum_identity_score
        elif str(method).lower() == 'all':
            Identity_score = [Average_identity_score, Minimum_identity_score, Maximum_identity_score]
    except ValueError:
        Identity_score = np.nan
    return Identity_score


'''
Copyright (c) 2022 Masayuki TAKAHASHI

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), 
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

import os
import sys
import warnings
import re
from chardet import detect
from chardet.universaldetector import UniversalDetector
import signal
import math
import numpy as np
import pandas as pd
import itertools as it
from scipy.cluster.hierarchy import distance, linkage, fcluster
try:
    import cupy as cp
    mempool = cp.get_default_memory_pool()
except:
    pass

def init_worker():
    """
    Keyboard interrapt function during parallel process. (private)

    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)

class nucleotide():
    """
    Single nucleotide definition.

    Attributes
    ----------
    base: str
        Nucleotide, One character description.
    Tm: float
        Tm value used for nucleotide sequence Tm value calculation by GC method.
    code: str
        binary code described nucleotide.
    """
    def __init__(self, args):
        """
        Parameters
        ----------
        args: str
            Nucleotide, One character description.
        """
        self.description = args.lower()
        if args in ("a", "A"):
            self.base = args.lower()
            self.Tm = 2.0
            self.code = '{:04b}'.format(int('1',16))
        elif args in ("t", "T"):
            self.base = args.lower()
            self.Tm = 2.0
            self.code = '{:04b}'.format(int('8',16))
        elif args in ("g", "G"):
            self.base = args.lower()
            self.Tm = 4.0
            self.code = '{:04b}'.format(int('2',16))
        elif args in ("c", "C"):
            self.base = args.lower()
            self.Tm = 4.0
            self.code = '{:04b}'.format(int('4',16))
        elif args in ("b", "B"):
            self.base = ("c", "g", "t")
            self.Tm = 3.33
            self.code = '{:04b}'.format(int('e',16))
        elif args in ("d", "D"):
            self.base = ("a", "g", "t")
            self.Tm = 2.67
            self.code = '{:04b}'.format(int('b',16))
        elif args in ("h", "H"):
            self.base = ("a", "c", "t")
            self.Tm = 2.67
            self.code = '{:04b}'.format(int('d',16))
        elif args in ("k", "K"):
            self.base = ("g", "t")
            self.Tm = 3.0
            self.code = '{:04b}'.format(int('a',16))
        elif args in ("m", "M"):
            self.base = ("a", "c")
            self.Tm = 3.0
            self.code = '{:04b}'.format(int('5',16))
        elif args in ("n", "N"):
            self.base = ("a", "c", "g", "t")
            self.Tm = 3.0
            self.code = '{:04b}'.format(int('f',16))
        elif args in ("r", "R"):
            self.base = ("a", "g")
            self.Tm = 3.0
            self.code = '{:04b}'.format(int('3',16))
        elif args in ("s", "S"):
            self.base = ("c", "g")
            self.Tm = 4.0
            self.code = '{:04b}'.format(int('6',16))
        elif args in ("v", "V"):
            self.base = ("a", "c", "g")
            self.Tm = 3.33
            self.code = '{:04b}'.format(int('7',16))
        elif args in ("w", "W"):
            self.base = ("a", "t")
            self.Tm = 2.0
            self.code = '{:04b}'.format(int('9',16))
        elif args in ("x", "X"):
            self.base = args.lower()
            self.Tm = 0.0
            self.code = '{:04b}'.format(int('0',16))
        elif args in ("y", "Y"):
            self.base = ("c", "t")
            self.Tm = 3.0
            self.code = '{:04b}'.format(int('c',16))
        else:
            self.base = ("a", "c", "g", "t")
            self.Tm = 3.0
            self.code = '{:04b}'.format(int('f',16))
        if args in ("a", "g", "r", "A", "G", "R"):
            self.type = "Pr"
            self.type_code = "01"
        elif args in ("c", "t", "y", "C", "T", "Y"):
            self.type = "Py"
            self.type_code = "10"
        elif args in ("n", "N"):
            self.type = "nt"
            self.type_code = "00"
        else:
            self.type = "Unknown"
            self.type_code = "00"
    def __repr__(self):
        return self.description
    def __str__(self):
        return self.description

class nucleotide_sequence(str):
    """
    Nucleotide sequence definition.

    Attributes
    ----------
    string: str
        Input character
    description: str
        Nuclotide sequence.
    sequence_length: int
        Nucleotide sequence length (nt).

    """
    def __init__(self, sequence):
        """
        Parameters
        ----------
        sequence: str
            Nucleotide sequence.
        """
        self.string = sequence
        self.description = sequence.upper()
        self.sequence_length = len(sequence)
    def __repr__(self):
        if len(self.description) > 100:
            return "<Sequence; [ {0} ... {1} ], Sequence length; {2}>".format(self.description[0:25], self.description[(len(self.description) - 25):len(self.description)], self.sequence_length)
        else:
            return "<Sequence; [ {0} ], Sequence length; {1}>".format(self.description, self.sequence_length)
    def __str__(self):
        return self.string
    def Nucleotide_composition(self):
        """
        Calculate the nucleotide composition.
        
        """
        return {"A":round(self.description.count("A")/(self.sequence_length - self.description.count("X")), 3), "T":round(self.description.count("T")/(self.sequence_length - self.description.count("X")), 3), "C":round(self.description.count("C")/(self.sequence_length - self.description.count("X")), 3), "G":round(self.description.count("G")/(self.sequence_length - self.description.count("X")), 3), "Others":round(((self.sequence_length - self.description.count("X")) - self.description.count("A") - self.description.count("T") - self.description.count("G") - self.description.count("C"))/(self.sequence_length - self.description.count("X")), 3)}
    def Decompress(self):
        """
        Generate the list of nucleotide sequences converted from a wobble base pair to ATGC from a sequence that has a wobble base pair.

        Parameters
        ----------
        None

        Returns
        ------
        sequences: list
        
        """
        return [str("".join(ntcomb)).upper() for ntcomb in it.product(*[nucleotide(nt).base for nt in list(self.description)])]
    def complementary_sequence(self):
        """
        Convert an input sequence to a complementary sequence.

        Parameters
        ----------
        None

        Returns
        -------
        New_sequences: str
            The complementary sequence of the sequence in the object.
        """
        subdict = {"a":"t", "t":"a", "g":"c", "c":"g", "b":"v", "d":"h", "h":"d", "k":"m", "m":"k", "r":"y", "s":"s", "v":"b", "w":"w", "y":"r", "n":"n", "x":"x"}
        sequence = str(self.description)[::-1].lower()
        new_sequences = "".join([subdict[nt] for nt in list(sequence)])
        new_sequences = new_sequences.upper()
        return new_sequences
    def calculate_score(self, seq):
        '''
        Calculate the similarity between two sequences. The higher the value is, the more similar the two sequences are.

        Parameters
        ----------
        seq: str or nucleotide_sequence

        Returns
        -------
        Similarity_score: float
        '''
        seq1 = self.description
        if type(seq) is nucleotide_sequence:
            seq2 = seq.description
        elif type(seq) is str:
            seq2 = seq
        elif len(re.sub(r"[A-Za-z]|\n", "", seq)) == 0:
            seq2 = str(seq)
        else:
            print("Input sequence should be class <str>, <nucleotide_sequence> or str-like object.")
            return None
        if len(seq1) == len(seq2):
            s1 = np.array([list(nucleotide(nt).code) + list(map(lambda x:float(x)/2, list(nucleotide(nt).type_code))) for nt in list(seq1)], dtype = float).astype(np.float32)
            s2 = np.array([list(nucleotide(nt).code) + list(map(lambda x:float(x)/2, list(nucleotide(nt).type_code))) for nt in list(seq2)], dtype = float).astype(np.float32)
            score = np.sum(np.max(s1 * s2, axis = 1))
            score = score/len(seq1)
            return score
        else:
            print("The length of sequences compared should be corresponded to each other.")
            return None
    def calculate_flexibility(self, detailed = False):
        '''
        Calculate the score for degeneracy.

        Parameters
        ----------
        detailed: bool
            The score when specifying True will be more accurate. (default: False)

        Returns
        -------
        Flexibility: float
            The higher the score is, the more wobble base pairs the sequence has.

        '''
        sequence = self.description
        if detailed:
            if (type(sequence) == list) | (type(sequence) == tuple):
                each_flexibility = [(seq.count("N") * 1 + seq.count("V") * 0.9902 + seq.count("B") * 0.9901 + seq.count("D") * 0.9900 + seq.count("H") * 0.9899 + seq.count("S") * 0.8502 + seq.count("K") * 0.8501 + seq.count("M") * 0.8500 + seq.count("W") * 0.8499 + seq.count("R") * 0.7501 + seq.count("Y") * 0.7499)/len(seq) if len(re.sub(r"[ATGC]", "", seq)) != 0 else 0.25/len(seq) for seq in sequence]
                flexibility = np.array(each_flexibility).prod() ** (1/len(sequence))
            else:
                flexibility = (sequence.count("N") * 1 + sequence.count("V") * 0.9902 + sequence.count("B") * 0.9901 + sequence.count("D") * 0.9900 + sequence.count("H") * 0.9899 + sequence.count("S") * 0.9502 + sequence.count("K") * 0.9501 + sequence.count("M") * 0.9500 + sequence.count("W") * 0.9499 + sequence.count("R") * 0.7501 + sequence.count("Y") * 0.7499)/len(sequence)
            return float(flexibility)
        else:
            if (type(sequence) == list) | (type(sequence) == tuple):
                each_flexibility = [(len(re.sub(r"[ATGCRY]", "", seq))+(len(seq)-len(re.sub(r"[RY]", "", seq)))*0.75)/len(seq) if len(re.sub(r"[ATGC]", "", seq)) != 0 else 0.25/len(seq) for seq in sequence]
                flexibility = np.array(each_flexibility).prod() ** (1/len(sequence))
            else:
                flexibility = (len(re.sub(r"[ATGCRY]", "", sequence))+(len(sequence)-len(re.sub(r"[RY]", "", sequence)))*0.75)/len(sequence)
            return float(flexibility)
    def array_diff(self, arr1, arr2, lower, upper, logical = False, separation_key = []):
        '''
        Calculate the difference between two arrays.

        Parameters
        ----------
        arr1: int, float, list, tuple, numpy.ndarray
        arr2: int, float, list, tuple, numpy.ndarray
        lower: int, float
        upper: int, float
        logical: bool
        separation_key: int, float, list of int, list of float

        Returns
        -------
        Result: list
            A list containing the subtract value is returned at default settings.
            When the logical is True, it returns a logical value.
        
        Notes
        -----
        Subtract arr1 from arr2.
        Subtraction across Separation_key is NOT performed. 
        For example, arr1 is [1, 1000, 10000], arr2 is [50, 1100, 10150], lower is 40 and upper is 400.
        The answer [49, 100, 150] will be obtained when the Separation key is the default value (default: −1).
        However, when the Separation key is 1050, [49, 150] will be obtained.

        '''
        iterable_or_not1 = np.any([((type(arr_comp) is list) | (type(arr_comp) is tuple) | (type(arr_comp) is np.ndarray)) for arr_comp in arr1])
        iterable_or_not2 = np.any([((type(arr_comp) is list) | (type(arr_comp) is tuple) | (type(arr_comp) is np.ndarray)) for arr_comp in arr2])
        if iterable_or_not1 & iterable_or_not2:
            if separation_key == []:
                separation_key = [[-1] for i in range(len(list(zip(arr1, arr2))))]
            elif (type(separation_key) is int) | (type(separation_key) is float):
                separation_key = [[separation_key] for i in range(len(list(zip(arr1, arr2))))]
            diff_ans = [np.array([b - a if ((pd.notna(a) & pd.notna(b)) & (np.logical_not(np.any((a <= np.array(sk)).astype(np.bool) & (np.array(sk) <= b).astype(np.bool))))) else np.nan for a in array1 for b in array2]) if (np.any(pd.notna(array1)) & np.any(pd.notna(array2))) else np.nan for array1, array2, sk in zip(arr1, arr2, separation_key)]
        elif iterable_or_not1 & np.logical_not(iterable_or_not2):
            if separation_key == []:
                separation_key = [[-1] for i in range(len(list(it.product(arr1, [arr2]))))]
            elif (type(separation_key) is int) | (type(separation_key) is float):
                separation_key = [[separation_key] for i in range(len(list(zip(arr1, arr2))))]
            diff_ans = [np.array([b - a if ((pd.notna(a) & pd.notna(b)) & (np.logical_not(np.any((a <= np.array(sk)).astype(np.bool) & (np.array(sk) <= b).astype(np.bool))))) else np.nan for a in array[0] for b in array[1]]) if (np.any(pd.notna(array[0])) & np.any(pd.notna(array[1]))) else np.nan for array, sk in zip(it.product(arr1, [arr2]), separation_key)]
        elif np.logical_not(iterable_or_not1) & iterable_or_not2:
            if separation_key == []:
                separation_key = [[-1] for i in range(len(list(it.product([arr1], arr2))))]
            elif (type(separation_key) is int) | (type(separation_key) is float):
                separation_key = [[separation_key] for i in range(len(list(zip(arr1, arr2))))]
            diff_ans = [np.array([b - a if ((pd.notna(a) & pd.notna(b)) & (np.logical_not(np.any((a <= np.array(sk)).astype(np.bool) & (np.array(sk) <= b).astype(np.bool))))) else np.nan for a in array[0] for b in array[1]]) if (np.any(pd.notna(array[0])) & np.any(pd.notna(array[1]))) else np.nan for array, sk in zip(it.product([arr1], arr2), separation_key)]
        else:
            if separation_key == []:
                separation_key = [-1]
            elif (type(separation_key) is int) | (type(separation_key) is float):
                separation_key = [separation_key]
            diff_ans = np.array([b - a if ((pd.notna(a) & pd.notna(b)) & (np.logical_not(np.any((a <= np.array(separation_key)).astype(np.bool) & (np.array(separation_key) <= b).astype(np.bool))))) else np.nan for a in arr1 for b in arr2])
            if logical:
                return np.any((diff_ans > lower) & (diff_ans < upper))
            else:
                return diff_ans[(diff_ans > lower) & (diff_ans < upper)]
        if logical:
            res = np.all([(np.any((ans > lower) & (ans < upper)) & np.all((ans > lower) | (ans <= 0) | (np.isnan(ans)))) if np.any(pd.notna(ans)) else True for ans in diff_ans])
        else:
            res = [ans[(ans > lower) & (ans < upper)] if np.any(pd.notnull(ans)) else np.array([]) for ans in diff_ans]
        return res
    def calculate_Tm_value(self, conc = 0.5, Na_conc = 0.05, mode = 'min'):
        '''
        Calculate the Tm value. 

        Parameters
        ----------
        conc: float
            The concentration of a primer when PCR amplification is performed. (default: 0.5 picomole/microliter)
        Na_conc: float
            The concentration of sodium when PCR amplification is performed. (default: 0.05 mole/liter)
        mode: str
            Choose a method from 'min', 'max', 'average', or ''. (default: min)

        Returns
        -------
        Tm_value: numpy.float or list

        '''
        Decompressed_sequence = self.Decompress()
        Parameter = {"AA": [-9.1, -0.0240], "TT": [-9.1, -0.0240], "AT": [-8.6, -0.0239], "TA": [-6.0, -0.0169], \
            "CA": [-5.8, -0.0129], "TG": [-5.8, -0.0129], "GT": [-6.5, -0.0173], "AC": [-6.5, -0.0173], \
            "CT": [-7.8, -0.0208], "AG": [-7.8, -0.0208], "GA": [-5.6, -0.0135], "TC": [-5.6, -0.0135], \
            "CG": [-11.9, -0.0278], "GC": [-11.1, -0.0267], "GG": [-11.0, -0.0266], "CC": [-11.0, -0.0266]}
        if len(Decompressed_sequence[0]) > 17:
            Enthalpy = [np.sum([Parameter[sequence[i:(i + 2):]][0] for i in range(len(sequence) - 1)]) for sequence in Decompressed_sequence]
            Entropy = [np.sum([Parameter[sequence[i:(i + 2):]][1] for i in range(len(sequence) - 1)]) for sequence in Decompressed_sequence]
            Tm_value = [round(((dH / (-0.0108 + dS + 0.001987 * math.log((conc / 1000000) / 4))) - 273.15 + 16.6 * math.log10(Na_conc)), 1) for dH, dS in zip(Enthalpy, Entropy)]
        else:
            Tm_value = [round(np.sum([nucleotide(nt).Tm for nt in list(sequence)]), 1) for sequence in Decompressed_sequence]
        if mode == 'min':
            Result = np.min(Tm_value)
        elif mode == 'max':
            Result = np.max(Tm_value)
        elif mode == 'average':
            Result = np.mean(Tm_value)
        else:
            Result = Tm_value
        return Result
    def search_position(self, evaluating_sequence, allowance = 0, interval_distance = 0, Match_rate = 0.0, circularDNAtemplate = False, IgnoreLowQualityRegion = True):
        '''
        Search an annealing site position .

        Parameters
        ----------
        evaluating_sequence: str or nucleotide_sequence
            The sequence for which you want to identify the annealing site position in a template sequence.

        Returns
        -------
        Start_position: dict
            The annealing site position between an evaluated sequence and a template (self) sequence.
            The absolute value of a position number indicates the onset point of the annealing site of a template sequence.
            e.g., If the "evaluating_sequence" is "ATGC", {"ATGC": [10, −35]} will be obtained.\n
                Template : ttggaatgagATGCtgtgaacagtcgtatatacgcGCATcgagattacgctattcgcgcggcg\n

        '''
        input_sequence = self.description
        if circularDNAtemplate:
            input_sequence = str(input_sequence) + str(input_sequence)[0:int(len(evaluating_sequence) - 1):]
        if IgnoreLowQualityRegion:
            IGNORE = math.ceil(np.max([(len(str(evaluating_sequence)) - allowance) / 2, 2]))
            IGNORE = r"".join(["N" for n in range(IGNORE)]) + r'+'
            IGNORE = re.findall(IGNORE, input_sequence)
            IGNORE.sort(reverse = True)
            for igr in IGNORE:
                input_sequence = re.sub(igr, igr.replace("N", "X"), input_sequence)
        if type(evaluating_sequence) is nucleotide_sequence:
            evaluating_sequence = evaluating_sequence.description
        encode_table = {"a":'1', "t":'8', "g":'2', "c":'4', "b":'e', "d":'b', "h":'d', "k":'a', "m":'5', "r":'3', "s":'6', "v":'7', "w":'9', "y":'c', "n":'f', "x":'0'}
        encode_table = str.maketrans(encode_table)
        if len(input_sequence) > len(evaluating_sequence):
            translate_sequence = np.array([list('{:04b}'.format(int(n,16))) for n in list(input_sequence.lower().translate(encode_table)) if re.sub(r'[0-9a-f]', '', n) == ''], dtype = int)
        else:
            print("\nInput data error in search homologous position.\n")
            sys.exit()
        translate_sequence_matrix = np.empty(((len(translate_sequence) - len(evaluating_sequence) + 1), 0), np.uint8)
        for i in range(len(evaluating_sequence) - 1):
            translate_sequence_matrix = np.concatenate([translate_sequence_matrix.astype(np.uint8), translate_sequence[i:(i - len(evaluating_sequence) + 1):].astype(np.uint8)], axis = 1)
        translate_sequence_matrix = np.concatenate([translate_sequence_matrix, translate_sequence[(len(evaluating_sequence) - 1)::]], axis = 1).astype(np.uint8)
        translate_evaluating_sequence = np.array([list('{:04b}'.format(int(n,16))) for n in list(evaluating_sequence.lower().translate(encode_table))], dtype = int)
        if "cupy" in sys.modules:
            try:
                translate_sequence_matrix_GPU = cp.asarray((translate_sequence_matrix.T).astype(np.uint8))
                translate_evaluating_sequence_GPU = cp.asarray(translate_evaluating_sequence.flatten().astype(np.uint8))
                score_list_preparation_GPU = cp.zeros(shape = (translate_sequence_matrix_GPU.shape[1], ), dtype = cp.uint8)
                cp.dot(translate_evaluating_sequence_GPU, translate_sequence_matrix_GPU, out = score_list_preparation_GPU)
                score_list_preparation = cp.asnumpy(score_list_preparation_GPU)
                cp.dot(translate_evaluating_sequence_GPU, cp.flipud(translate_sequence_matrix_GPU), out = score_list_preparation_GPU)
                complementary_score_list_preparation = cp.asnumpy(score_list_preparation_GPU)
                del score_list_preparation_GPU, translate_sequence_matrix_GPU, translate_evaluating_sequence_GPU
                mempool.free_all_blocks()
            except:
                if 'translate_sequence_matrix_GPU' in locals(): del translate_sequence_matrix_GPU
                if 'translate_evaluating_sequence_GPU' in locals(): del translate_evaluating_sequence_GPU
                if 'score_list_preparation_GPU' in locals(): del score_list_preparation_GPU
                mempool.free_all_blocks()
                translate_sequence_matrix_CPU = (translate_sequence_matrix.T).astype(np.uint8)
                translate_evaluating_sequence_CPU = translate_evaluating_sequence.flatten().astype(np.uint8)
                score_list_preparation = np.dot(translate_evaluating_sequence_CPU, translate_sequence_matrix_CPU).astype(np.uint8)
                complementary_score_list_preparation = np.dot(translate_evaluating_sequence_CPU, np.flipud(translate_sequence_matrix_CPU)).astype(np.uint8)
        else:
            translate_sequence_matrix_CPU = (translate_sequence_matrix.T).astype(np.uint8)
            translate_evaluating_sequence_CPU = translate_evaluating_sequence.flatten().astype(np.uint8)
            score_list_preparation = np.dot(translate_evaluating_sequence_CPU, translate_sequence_matrix_CPU).astype(np.uint8)
            complementary_score_list_preparation = np.dot(translate_evaluating_sequence_CPU, np.flipud(translate_sequence_matrix_CPU)).astype(np.uint8)
        IDX = np.where(score_list_preparation.astype(np.float32) >= (len(evaluating_sequence) - allowance))[0]
        score_list = score_list_preparation.astype(np.float32) / len(evaluating_sequence)
        score_list[IDX] = [nucleotide_sequence(evaluating_sequence).calculate_score(input_sequence[id:(id + len(evaluating_sequence)):]) for id in IDX]
        score_list_criteria = list(np.where(np.array(score_list) >= ((len(evaluating_sequence) - allowance)/len(evaluating_sequence)))[0])
        IDX = np.where(complementary_score_list_preparation.astype(np.float32) >= (len(evaluating_sequence) - allowance))[0]
        complementary_score_list = complementary_score_list_preparation.astype(np.float32) / len(evaluating_sequence)
        complementary_score_list[IDX] = [nucleotide_sequence(evaluating_sequence).calculate_score(complementary_sequence(input_sequence[id:(id + len(evaluating_sequence)):])) for id in IDX]
        score_list = np.maximum(score_list, complementary_score_list)
        Start_positions = dict()
        Mismatch_Counter = 0
        Total_Counter = 0
        Maximum_position_list_size = 12
        Maximum_combination_size = 30000
        Position_list = list(np.where(np.array(score_list) >= ((len(evaluating_sequence) - allowance)/len(evaluating_sequence)))[0])
        score_list_criteria = np.array([set(score_list_criteria) >= set([Position]) for Position in Position_list])
        Complementary_criteria = list(np.where(np.array(complementary_score_list) >= ((len(evaluating_sequence) - allowance)/len(evaluating_sequence)))[0])
        Complementary_criteria = np.array([set(Complementary_criteria) >= set([Position]) for Position in Position_list])
        Position_list = np.array(Position_list)
        if len(Position_list) != 0:
            Palindrome_sequence_list = {}
            Palindrome_position = Position_list[score_list_criteria & Complementary_criteria]
            [Palindrome_sequence_list.update({input_sequence[Ppos:(Ppos + len(evaluating_sequence)):]: Palindrome_sequence_list[input_sequence[Ppos:(Ppos + len(evaluating_sequence)):]] + [Ppos]}) if (input_sequence[Ppos:(Ppos + len(evaluating_sequence)):] in Palindrome_sequence_list.keys()) else Palindrome_sequence_list.update({input_sequence[Ppos:(Ppos + len(evaluating_sequence)):]: [Ppos]}) for Ppos in Palindrome_position]
        else:
            Palindrome_sequence_list = {}
        if np.any(Complementary_criteria):
            Position_list[Complementary_criteria] = Position_list[Complementary_criteria] * (-1)
        if len(Position_list) != 0:
            score_list = np.array(score_list)[np.abs(Position_list)]
            split_sequences = np.array([input_sequence[Position:(Position + len(evaluating_sequence)):] if Position >= 0 else complementary_sequence(input_sequence[np.abs(Position):(np.abs(Position) + len(evaluating_sequence)):]) for Position in Position_list])
            duplicate_index = [split_sequences == seq for seq in sorted(list(set(split_sequences)))]
            Grouping_position_list = [list(Position_list[didx]) for didx in duplicate_index]
            Grouping_split_sequences = [list(set(split_sequences[didx])) for didx in duplicate_index]
            if len(Grouping_position_list) > Maximum_position_list_size:
                Position_combinations = [tuple([sorted(list(it.chain.from_iterable(Grouping_position_list)), key = abs)])]
            else:
                Group_combination_numbers = [[np.array(Grouping_combination_number) for Grouping_combination_number in list(it.combinations(range(len(Grouping_position_list)), i))] for i in range(1, len(Grouping_position_list) + 1)]
                Position_combinations = list(it.chain.from_iterable([[tuple(sorted(it.chain.from_iterable(np.array(Grouping_position_list, dtype = object)[Gcn].tolist()), key = abs)) for Gcn in Group_combination_number] for Group_combination_number in Group_combination_numbers]))
                Split_sequences_combinations = list(it.chain.from_iterable([[tuple(it.chain.from_iterable(np.array(Grouping_split_sequences, dtype = object)[Gcn].tolist())) for Gcn in Group_combination_number] for Group_combination_number in Group_combination_numbers]))
                Combination_in_allowance = [len(re.sub('[ATGC]', '', make_wobble(*Split_sequences_combination)[0])) <= np.max([allowance, np.max([len(re.sub('[ATGC]', '', seq)) for seq in Split_sequences_combination])]) for Split_sequences_combination in Split_sequences_combinations]
                Position_combinations = [Position_combinations[i] for i in range(len(Position_combinations)) if Combination_in_allowance[i]]
                if len(Position_combinations) > Maximum_combination_size:
                    Position_combinations = np.array(tuple([sorted(list(it.chain.from_iterable(Grouping_position_list)), key = abs)]), dtype = object)
            Criteria = np.array([[np.count_nonzero(np.array(posc) == pos) > 0 for pos in Position_list] for posc in Position_combinations])
            score_list = [np.array(score_list)[cr] for cr in Criteria]
            del split_sequences
            for i in range(len(Position_combinations)):
                for in_allowance in range(allowance + 1):
                    Position_index = np.array(Position_combinations[i])[np.where(np.array(score_list[i]) >= ((len(evaluating_sequence) - in_allowance)/len(evaluating_sequence)))[0].tolist()]
                    if np.any([np.all([np.all(np.diff(np.sort(np.abs(Position_index))) >= interval_distance)]), (len(Position_index) == 1)]):
                        extract_input_sequences = [input_sequence[Position:(Position + len(evaluating_sequence)):] if Position >= 0 else complementary_sequence(input_sequence[np.abs(Position):(np.abs(Position) + len(evaluating_sequence)):]) for Position in Position_index]
                        extract_input_sequences = [nucleotide_sequence(extract_input_sequence).Decompress() if len(re.sub('[ATGC]', '', extract_input_sequence)) > 0 else [extract_input_sequence] for extract_input_sequence in extract_input_sequences]
                        extract_input_sequence_score = [np.argmax([nucleotide_sequence(decompressed_seq).calculate_score(evaluating_sequence) for decompressed_seq in extract_input_sequence]) for extract_input_sequence in extract_input_sequences]
                        extract_input_sequences = [str(np.array(extract_input_sequences[i])[extract_input_sequence_score[i]]) for i in range(len(extract_input_sequences))]
                        seq = make_wobble(*([evaluating_sequence] + extract_input_sequences))
                        Position_index = Position_list[np.array([nucleotide_sequence(seq[0]).calculate_score(input_sequence[Position:(Position + len(evaluating_sequence)):]) == 1 if Position >= 0 else nucleotide_sequence(seq[0]).calculate_score(complementary_sequence(input_sequence[np.abs(Position):(np.abs(Position) + len(evaluating_sequence)):])) == 1 for Position in Position_list])].astype(object)
                        Start_position = {seq[0]:Position_index} if np.all([(nucleotide_sequence(seq[0]).calculate_flexibility() <= (nucleotide_sequence(evaluating_sequence).calculate_flexibility() + (in_allowance/len(evaluating_sequence)))), (nucleotide_sequence(seq[0]).calculate_flexibility() < 1.0), len(Position_index) > 0]) else None
                        Start_positions.update({list(Start_position.keys())[0]:np.array(sorted(list(set(Start_positions[list(Start_position.keys())[0]].tolist() + Start_position[list(Start_position.keys())[0]].tolist())), key = abs), dtype = object)} if (set(Start_position.keys()) <= set(Start_positions.keys())) else Start_position) if Start_position is not None else Start_positions.update()
                        Mismatch_Counter = Mismatch_Counter if np.any([np.all([np.all(np.diff(np.sort(np.abs(Position_index))) >= interval_distance)]), (len(Position_index) == 1)]) else Mismatch_Counter + 1
                    else:
                        Mismatch_Counter += 1
                    Total_Counter += 1
        else:
            pass
        Mismatch_rate = Mismatch_Counter / Total_Counter if ((allowance > 0) & (Total_Counter > 0)) else 0
        if len(Palindrome_sequence_list) != 0:
            Start_positions = {Key: Position.tolist() + [pp * (-1) if set(Position) >= set([pp]) else pp for palseq, Ppos in Palindrome_sequence_list.items() if ((nucleotide_sequence(Key).calculate_score(palseq) == 1.0) & (nucleotide_sequence(Key).calculate_score(complementary_sequence(palseq)) == 1.0)) for pp in Ppos] for Key, Position in Start_positions.items()}
            Start_positions = {Key: np.array(sorted(sorted(list(set(Position))), key = abs), dtype = object) for Key, Position in Start_positions.items()}
        Start_positions = {Key:Position for Key, Position in Start_positions.items() if np.all(np.diff(np.sort(np.abs(Position))) >= interval_distance)}
        if (len(Start_positions) != 0) & ((1 - Mismatch_rate) >= Match_rate):
            return Start_positions
        else:
            return None
    def PCR_amplicon(self, forward, reverse, allowance = 0, Single_amplicon = True, Sequence_Only = True, amplicon_size_limit = 10000, Warning_ignore = False, circularDNAtemplate = False, IgnoreLowQualityRegion = True):
        '''
        In silico PCR.

        Parameters
        ----------
        forward: str or nucleotide_sequence
            Forward primer sequence (required).
        reverse: str or nucleotide_sequence
            Reverse primer sequence (required).
        allowance: int
            The acceptable mismatch number. (default: 0)
        Single_amplicon: bool
            All amplicons that will be amplified by an input primer set are outputed as list when False is specified. (default: True)
        Sequence_Only: bool
            The start and end positions of an amplicon in template sequence are outputed with the amplicon sequence when False is specified. (default: True)
        Warning_ignore: bool
            Show all warnings if specify True. (default: False)
        circularDNAtemplate: bool
            Specify True if the input sequence is circular DNA. (default: False)

        Returns
        -------
        PCR_amplicon: str or list

        '''
        if Warning_ignore:
            warnings.simplefilter('ignore')
        template = self.description
        Original_template_length = len(str(template))
        if circularDNAtemplate:
            template = str(template) + str(template)[0:int(np.min([amplicon_size_limit, Original_template_length * 0.1])):]
        if IgnoreLowQualityRegion:
            IGNORE = math.ceil(np.max([(np.min([len(str(forward)), len(str(reverse))]) - allowance) / 2, 2]))
            IGNORE = r"".join(["N" for n in range(IGNORE)]) + r'+'
            IGNORE = re.findall(IGNORE, template)
            IGNORE.sort(reverse = True)
            for igr in IGNORE:
                template = re.sub(igr, igr.replace("N", "X"), template)
        template = nucleotide_sequence(template)
        PCR_amplicon = None
        if type(forward) is not nucleotide_sequence:
            forward = nucleotide_sequence(forward)
        if type(reverse) is not nucleotide_sequence:
            reverse = nucleotide_sequence(reverse)
        forward_position = template.search_position(evaluating_sequence = forward.description, allowance = allowance, interval_distance = 0, Match_rate = 0.0, IgnoreLowQualityRegion = False)
        reverse_position = template.search_position(evaluating_sequence = reverse.complementary_sequence(), allowance = allowance, interval_distance = 0, Match_rate = 0.0, IgnoreLowQualityRegion = False)
        if (forward_position is not None) & (reverse_position is not None):
            forward_position = sorted(sorted(list(set(list(it.chain.from_iterable([v for v in forward_position.values()]))))), key = abs)
            reverse_position = sorted(sorted(list(set(list(it.chain.from_iterable([v for v in reverse_position.values()]))))), key = abs)
            if Sequence_Only:
                Amplicon_size = [self.array_diff([f], reverse_position, lower = 0, upper = amplicon_size_limit, logical = False, separation_key = []) + reverse.sequence_length if f >= 0 else self.array_diff([f], reverse_position, lower = 0, upper = amplicon_size_limit, logical = False, separation_key = []) + forward.sequence_length for f in forward_position]
                PCR_amplicon = [template[int(f):int(f+a):] if f >= 0 else template[int(np.abs(f) + forward.sequence_length - a):int(np.abs(f) + forward.sequence_length):] for f, amp in zip(forward_position, Amplicon_size) for a in amp]
                PCR_amplicon = list(set([amp for amp in PCR_amplicon if ((len(amp) <= amplicon_size_limit) & (len(amp) != 0) & (amp.find("X") < 0))]))
            else:
                Amplicon_size = [self.array_diff([f], reverse_position, lower = 0, upper = amplicon_size_limit, logical = False, separation_key = []) + reverse.sequence_length if f >= 0 else self.array_diff([f], reverse_position, lower = 0, upper = amplicon_size_limit, logical = False, separation_key = []) + forward.sequence_length for f in forward_position]
                PCR_amplicon = [(int(f + 1), int(f + a), template[int(f):int(f+a):], ) if f >= 0 else (int(f - forward.sequence_length), int(f - forward.sequence_length + a - 1), template[int(np.abs(f) + forward.sequence_length - a):int(np.abs(f) + forward.sequence_length):], ) for f, amp in zip(forward_position, Amplicon_size) for a in amp]
                PCR_amplicon = [(f, r, amp, ) for f, r, amp in PCR_amplicon if ((len(amp) <= amplicon_size_limit) & (len(amp) != 0) & (amp.find("X") < 0))]
                PCR_amplicon = [(f - Original_template_length, r, amp, ) if f >= Original_template_length else (f, r, amp, ) for f, r, amp in PCR_amplicon]
                PCR_amplicon = [(f, r - Original_template_length, amp, ) if r >= Original_template_length else (f, r, amp, ) for f, r, amp in PCR_amplicon]
                PCR_amplicon = sorted(list(set(PCR_amplicon)))
        if ((PCR_amplicon is None) | (PCR_amplicon == [])):
            PCR_amplicon = None
        elif Single_amplicon:
            PCR_amplicon = PCR_amplicon[0]
        else:
            PCR_amplicon = PCR_amplicon
        if ((forward_position is None) | (reverse_position is None) | (PCR_amplicon is None)):
            warnings.warn("\nNo amplicon was obtained.\n", stacklevel = 2)
        elif Single_amplicon & ((len(PCR_amplicon) > 1) & np.logical_not(((type(PCR_amplicon) is str)) | (type(PCR_amplicon) is tuple))):
            warnings.warn("\nTwo or more fragments will be amplified.\n", stacklevel = 2)
        else:
            pass
        return PCR_amplicon

def complementary_sequence(sequence):
    """
    Convert an input sequence to a complementary sequence.

    Parameters
    ----------
    sequence: str, nucleotide_sequence, list, tuple

    Returns
    ------
    New_sequences: str
        The complementary sequence of the sequence in the object.
    """
    subdict = {"a":"t", "t":"a", "g":"c", "c":"g", "b":"v", "d":"h", "h":"d", "k":"m", "m":"k", "r":"y", "s":"s", "v":"b", "w":"w", "y":"r", "n":"n", "x":"x"}
    if (type(sequence) is pd.DataFrame) | (type(sequence) is pd.Series) | (type(sequence) is list) | (type(sequence) is tuple):
        if (type(sequence) is pd.DataFrame) | (type(sequence) is pd.Series):
            index = sequence.index
            sequences = list(sequence.values)
        sequences = list(sequence)
        new_sequences = []
        for sequence in sequences:
            sequence = str(sequence)[::-1].lower()
            sequence = "".join([subdict[nt] for nt in list(sequence)])
            new_sequences.append(sequence.upper())
        if (type(sequence) is pd.DataFrame) | (type(sequence) is pd.Series):
            new_sequences = pd.Series(new_sequences, index = index)
    else:
        sequence = str(sequence)[::-1].lower()
        new_sequences = "".join([subdict[nt] for nt in list(sequence)])
        new_sequences = new_sequences.upper()
    return new_sequences

def calculate_Tm_value(sequence, conc = 0.5, Na_conc = 0.05, mode = 'min'):
    '''
    Calculate the Tm value. 

    Parameters
    ----------
    sequence: str, nucleotide_sequence, list, tuple or numpy.ndarray
        The sequence for which you would like to know the Tm value, or its list, tuple, or ndarray. (required)
    conc: float
        The concentration of a primer when PCR amplification is performed. (default: 0.5 picomole/microliter)
    Na_conc: float
        The concentration of sodium when PCR amplification is performed. (default: 0.05 mole/liter)
    mode: str
        Choose a method from 'min', 'max', 'average', or ''. (default: min)

    Returns
    -------
    Tm_value: numpy.float or list

    '''
    if (type(sequence) is list) | (type(sequence) is tuple):
        sequences = list(sequence)
    elif (type(sequence) is np.ndarray):
        sequences = sequence.tolist()
    elif (type(sequence) is str) | (type(sequence) is nucleotide_sequence):
        sequences = [str(sequence)]
    else:
        try:
            raise TypeError
        except:
            print("TypeError: ")
    Decompressed_sequences = [[str("".join(ntcomb)).upper() for ntcomb in it.product(*[nucleotide(nt).base for nt in list(seq)])] for seq in sequences]
    Tm_value_list = []
    Parameter = {"AA": [-9.1, -0.0240], "TT": [-9.1, -0.0240], "AT": [-8.6, -0.0239], "TA": [-6.0, -0.0169], \
        "CA": [-5.8, -0.0129], "TG": [-5.8, -0.0129], "GT": [-6.5, -0.0173], "AC": [-6.5, -0.0173], \
        "CT": [-7.8, -0.0208], "AG": [-7.8, -0.0208], "GA": [-5.6, -0.0135], "TC": [-5.6, -0.0135], \
        "CG": [-11.9, -0.0278], "GC": [-11.1, -0.0267], "GG": [-11.0, -0.0266], "CC": [-11.0, -0.0266]}
    for Decompressed_sequence in Decompressed_sequences:
        if len(Decompressed_sequence[0]) > 17:
            Enthalpy = [np.sum([Parameter[sequence[i:(i + 2):]][0] for i in range(len(sequence) - 1)]) for sequence in Decompressed_sequence]
            Entropy = [np.sum([Parameter[sequence[i:(i + 2):]][1] for i in range(len(sequence) - 1)]) for sequence in Decompressed_sequence]
            Tm_value = [round(((dH / (-0.0108 + dS + 0.001987 * math.log((conc / 1000000) / 4))) - 273.15 + 16.6 * math.log10(Na_conc)), 1) for dH, dS in zip(Enthalpy, Entropy)]
        else:
            Tm_value = [round(np.sum([nucleotide(nt).Tm for nt in list(sequence)]), 1) for sequence in Decompressed_sequence]
        if mode == 'min':
            Tm_value_list.append(np.min(Tm_value))
        elif mode == 'max':
            Tm_value_list.append(np.max(Tm_value))
        elif mode == 'average':
            Tm_value_list.append(np.mean(Tm_value))
        else:
            Tm_value_list.append(Tm_value)
    Result = dict(zip(sequences, Tm_value_list))
    Result = float(Result[list(Result.keys())[0]]) if (len(Result) == 1) & (type(Result[list(Result.keys())[0]]) is not list) else Result
    return Result

def read_sequence_file(input, format = "fasta", Feature = False):
    '''
    Read a sequence file. FASTA, Multi-FASTA, and GenBank sequence format files are accepted.

    Parameters
    ----------
    input: str
        File path. (required)
    format: str
        'fasta' or 'genbank' (default: fasta)
    Feature: bool
        The annotations are extracted if the GenBank format is input when True is specified. (default: False)

    Returns
    -------
    sequences: list
        A list will be returned as follows. [{'sequence name':sequence <nucleotide_sequence>}, input sequence number, average length of input sequence].
        In cases in which True is specified in a Feature argument, [{'sequence name':[sequence <nucleotide_sequence>, feature]}, input sequence number, average length of input sequence].

    '''
    with open(input, 'rb') as fin:
        detector = UniversalDetector()
        for line in fin:
            detector.feed(line)
            if detector.done:
                break
        detector.close()
        encoding = detector.result
    if encoding['encoding'] is None:
        with open(input, 'rb') as fin:
            f = fin.read()
            encoding = detect(f)
        del f
    with open(input, mode = "rt", encoding = encoding['encoding']) as fin:
        line = fin.readline()
        if line.startswith("LOCUS"):
            format = "genbank"
        if format.lower() == "fasta":
            name = ''
            seq = ''
            name_list = []
            seq_list = []
            while line:
                line = line.replace('\n','')
                if line.startswith('>'):
                    name = line[1::]
                    seq = ''
                else:
                    seq += line
                line = fin.readline()
                if not line or line.startswith('>'):
                    name_list.append(name)
                    seq_list.append(seq)
            seq_list = [nucleotide_sequence(re.sub('U', 'T', re.sub(r'[EFIJLOPQZ]', 'N', re.sub(r'[^A-Z]', '', seq.upper())))) for seq in seq_list]
            seq_dict = dict(zip(name_list,seq_list))
        elif format.lower() == "genbank":
            seq = ""
            gene_region = []
            gene_product = []
            SEQUENCE_Line = False
            FEATURES_Line = False
            GENE_Count = 0
            while len(line) > 0:
                if line.replace(" ", "").startswith("DEFINITION"):
                    name = line.replace("DEFINITION", "").replace("\n", "")
                    while True:
                        if not name.startswith(" "):
                            break
                        name = name[1::]
                if FEATURES_Line:
                    FEATURES_Key = ["C_region", "CDS", "D_segment", "exon", \
                                    "gene", "J_segment", "mat_peptide", "misc_feature", \
                                    "misc_RNA", "mRNA", "ncRNA", "N_region", "precursor_RNA", \
                                    "propeptide", "rRNA", "S_region", "sig_peptide", "tmRNA", \
                                    "transit_peptide", "tRNA", "V_region", "V_segment", "variation"]
                    if np.any([line.replace(" ", "").startswith(feature_key) for feature_key in FEATURES_Key]) & (re.search(r'[0-9]+\.+[^0-9]*[0-9]+', line) is not None):
                        GENE_Count += 1
                        gene_region.append(re.sub(r'['+r"|".join(FEATURES_Key)+r"|\n| |complement(|)]", "", line))
                    if line.replace(" ", "").startswith("/product="):
                        gene_product += ["" for n in range(GENE_Count - len(gene_product) - 1)]
                        PRODUCT = line.replace("/product=", "").replace('"', "").replace("\n", "")
                        while True:
                            if not PRODUCT.startswith(" "):
                                break
                            PRODUCT = PRODUCT[1::]
                        gene_product.append(PRODUCT)
                if SEQUENCE_Line:
                    seq += re.sub(r"[0-9| |\n|/]", "", line)
                if line.replace(" ", "").startswith("FEATURES"):
                    SEQUENCE_Line = False
                    FEATURES_Line = True
                if line.replace(" ", "").startswith("ORIGIN"):
                    SEQUENCE_Line = True
                    FEATURES_Line = False
                line = fin.readline()
            feature = {k:v for k, v in zip(gene_region, gene_product) if len(v) != 0}
            seq = re.sub('U', 'T', re.sub(r'[EFIJLOPQZ]', 'N', re.sub(r'[^A-Z]', '', seq.upper())))
            seq_list = [nucleotide_sequence(seq)]
            if 'name' not in locals():
                print("'{0}' is not Genbank file format. Reconfirm the file.\nWhen '--Annotaion' option is used, the input sequence file should be provided by genbank format.".format(input))
                sys.exit()
            if Feature:
                seq_dict = {name:[nucleotide_sequence(seq), feature]}
            else:
                seq_dict = {name:nucleotide_sequence(seq)}
        else:
            print("'{0}' format is not supported. Please convert '{0}' to 'fasta' or 'genbank'.".format(format))
            sys.exit()
    input_seq = len(seq_dict)
    Total_length = 0
    for inseq in seq_list:
        Total_length += inseq.sequence_length
    Average_length = Total_length/input_seq
    return [seq_dict, input_seq, Average_length]

def circularDNA(sequence, overlap = 0):
    """
    Convert linear DNA to circular DNA. (private)
    
    """
    new_sequence = str(sequence) + str(sequence)[0:int(overlap):]
    return new_sequence

def check_input_file(file_paths, circular = "all"):
    """
    Check input file. (private)
    
    """
    check_sequences = {}
    if type(file_paths) is str:
        file_paths = [file_paths]
    for file_path in file_paths:
        check_sequences.update(read_sequence_file(file_path)[0])
    if np.any([seq.sequence_length == 0 for name, seq in check_sequences.items()]):
        print("Some input sequence(s) doesn't contain sequence data.")
        sys.exit()
    if np.any([len(re.sub(r'[ATGC]', '', str(seq).upper()))/seq.sequence_length > 0.7 for name, seq in check_sequences.items()]) | np.any([len(re.sub(r'[ATGCBDHKMRSVWYNX]', '', str(seq).upper())) > 1 for name, seq in check_sequences.items()]):
        print("There is a possibility that Some input sequence(s) is not nucleotide sequence. Please check input sequence again.")
        sys.exit()
    if circular.lower() == "all":
        circular_DNA = [True for i in range(len(check_sequences))]
    elif circular.lower() == "individually":
        circular_DNA = [False for i in range(len(check_sequences))]
        for i in range(len(check_sequences)):
            check = ''
            while True:
                check = input("Is '{0}' circular DNA? (y/n/a-y[all yes after this sequence.]/a-n[all no after this sequence.]): ".format(str(list(check_sequences.keys())[i])))
                if (check.lower() == "all-yes") | (check.lower() == "a-y") | (check.lower() == "all-no") | (check.lower() == "a-n") |(check.lower() == "yes") | (check.lower() == "y") | (check.lower() == "no") | (check.lower() == "n"):
                    break
            if (check.lower() == "all-yes") | (check.lower() == "a-y"):
                circular_DNA[i:] = [True for i in range(len(check_sequences) - i)]
                break
            elif (check.lower() == "all-no") | (check.lower() == "a-n"):
                circular_DNA[i:] = [False for i in range(len(check_sequences) - i)]
                break
            elif (check.lower() == "yes") | (check.lower() == "y"):
                circular_DNA[i] = True
            elif (check.lower() == "no") | (check.lower() == "n"):
                circular_DNA[i] = False
    elif os.path.isfile(circular):
        circularDNAfile = []
        with open(circular, 'rt', encoding = "utf-8") as fin:
            line = fin.readline()
            while line:
                circularDNAfile.append(line)
                line = fin.readline()
        circular_DNA = [any([(n.replace(name, '').count('circularDNA') > 0) | (n.replace(name, '').count('circular') > 0) for n in circularDNAfile if ((n.count(name + " ") > 0) | (n.count(name + "\t") > 0) | (n.count(name + ",") > 0))]) for name in check_sequences.keys()]
    elif circular.lower() == "n/a":
        circular_DNA = [False for i in range(len(check_sequences))]
    else:
        print("Specify the file path, 'all', 'indivudually' or 'n/a' after circular argment.")
        sys.exit()
    circular_index = dict(zip(list(check_sequences.keys()), circular_DNA))
    return circular_index

def make_wobble(*args):
    '''
    Make a degenerate sequence from an input sequences list. (private)

    Parameters
    ----------
    args: str, unpacked list or unpacked tuple

    Returns
    -------
    New_sequence: list
        A list of a degenerate sequence and input sequences. [degenerate sequence, [Input sequences]]

    '''
    enc_submatrix = {"a":1, "t":8, "g":2, "c":4, "b":14, "d":11, "h":13, "k":10, "m":5, "r":3, "s":6, "v":7, "w":9, "y":12, "n":15, "x":0}
    dec_submatrix = {1:"a", 8:"t", 2:"g", 4:"c", 14:"b", 11:"d", 13:"h", 10:"k", 5:"m", 3:"r", 6:"s", 7:"v", 9:"w", 12:"y", 15:"n", 0:"x"}
    sequences = list(args)
    new_sequence = []
    nt_matrix = np.array([[enc_submatrix[nt.lower()] for nt in list(seq)] for seq in sequences]).T
    for i in range(nt_matrix.shape[0]):
        if np.any(nt_matrix[i] == 0):
            new_sequence.append(dec_submatrix[0])
        elif len(set(nt_matrix[i])) >= 9:
            new_sequence.append(dec_submatrix[15])
        else:
            bit_value = 0
            dec = []
            for n1 in set(nt_matrix[i]):
                bit_value += int(bin(n1)[2:])
            for n2 in str(bit_value):
                if int(n2) >= 1:
                    dec.append("1")
                else:
                    dec.append("0")
            dec = int("".join(dec),2)
            new_sequence.append(dec_submatrix[dec])
    new_sequence = "".join(new_sequence)
    result = [new_sequence.upper(), sequences]
    return result

def combination_basedon_correlation(df, number = 1, Correlation_threshold = 0.9):
    '''
    Generate iterator for combination of df.index based on a correlation calculated from values in every row. (private)

    Parameters
    ----------
    df: pandas.DataFrame.
    number: int
        The number of indexes contained in each combination. (default: 1)
    Correlation_threshold: float
        The indexes have the correlation score above this threshold are grouped, and two or more indexes are NEVER chosen from the indexes in same group. (default: 0.9)

    Returns
    -------
    Generator object

    '''
    try:
        df = df.fillna(0)
        if type(df.iloc[0,0]) is list:
            Maximum_element_number = [np.max(df.iloc[:, i].apply(lambda x:len(x) if type(x) is list else 1)) for i in range(df.shape[1])]
            new_df = pd.DataFrame([[[sorted(li)[j] if len(li) >= j + 1 else int(0) for j in range(n)] for li in df.iloc[:, i]] for i, n in enumerate(Maximum_element_number)], dtype = object).T
            new_df.index = df.index
            df = pd.DataFrame([list(it.chain.from_iterable(new_df.iloc[i])) for i in range(new_df.shape[0])], index = new_df.index)
        Cor_matrix = df.astype('float64').T.corr()
        Cor_matrix = np.nan_to_num(Cor_matrix)
        Cor_distance = distance.pdist(Cor_matrix)
        Cor_matrix[Cor_matrix < Correlation_threshold] = -10
        col_min_cor = [Cor_matrix[i, j] for i, j in enumerate(list(abs(Cor_matrix - Correlation_threshold).argmin(axis = 1)))]
        idx = [tuple([i, j]) for i, j in enumerate(list(abs(Cor_matrix - Correlation_threshold).argmin(axis = 1)))]
        row_min_idx = ([abs(num - Correlation_threshold) for num in col_min_cor]).index(min([abs(num - Correlation_threshold) for num in col_min_cor]))
        idx = idx[row_min_idx]
        Threshold = (distance.squareform(Cor_distance))[idx]
        Cor_linkage = linkage(Cor_distance, metric = "euclidean", method = "average")
        Group = fcluster(Cor_linkage, Threshold, criterion = "distance")
        Index = df.index
        Group_list = {}
        for i in set(Group):
            Group_list.update({i:list(Index[Group == i])})
        Group_iter = it.combinations(Group_list.keys(), number)
        for keys in Group_iter:
            GLE = [Group_list[key] for key in keys]
            for sequence_set in it.product(*GLE):
                yield sequence_set
    except AttributeError:
        print("First positional argument 'df' should be pandas.DataFrame.")
    except:
        raise
